#!/usr/bin/env python

# See the accompanying LICENSE file.

import os
import sys
import shlex
import glob
import re
import time
import zipfile
import tarfile

try:
    if not os.environ.get("APSW_FORCE_DISTUTILS"):
        from setuptools import setup, Extension, Command
    else:
        raise ImportError()
except ImportError:
    from distutils.core import setup, Extension, Command

from distutils.command import build_ext, build, sdist

##
## Do your customizations here or by creating a setup.cfg as documented at
## http://www.python.org/doc/2.5.2/dist/setup-config.html
##

include_dirs=['src']
library_dirs=[]
define_macros=[]
libraries=[]

# This includes the functionality marked as experimental in SQLite 3.
# Comment out the line to exclude them
define_macros.append( ('EXPERIMENTAL', '1') )

##
## End of customizations
##

# python 2 and 3 print equivalent
def write(*args):
    # py2 won't allow optional keyword arg on end, so work around it
    dest=sys.stdout
    if args[-1]==sys.stderr:
        dest=args[-1]
        args=args[:-1]
    dest.write(" ".join(args)+"\n")
    dest.flush()

py3=sys.version_info>=(3,0)

# ensure files are closed
def read_whole_file(name, mode):
    if sys.version_info<(2,4):
        if "r" in mode and "U" in mode:
            # python 2.3 returns file not found if "U" present!
            mode="".join([m for m in mode if m!="U"])
    f=open(name, mode)
    try:
        return f.read()
    finally:
        f.close()

def write_whole_file(name, mode, data):
    f=open(name, mode)
    try:
        f.write(data)
    finally:
        f.close()

# They keep messing with where files are in URI
def fixup_download_url(url):
    ver=re.search("3[0-9]{6}", url)
    if ver:
        ver=int(ver.group(0))
        if ver>=3071600:
            if ver>=3220000:
                year="2018"
            elif ver>=3160000:
                year="2017"
            elif ver>=3100000:
                year="2016"
            elif ver>=3080800:
                year="2015"
            elif ver>=3080300:
                year="2014"
            else:
                year="2013"
            if "/"+year+"/" not in url:
                url=url.split("/")
                url.insert(3, year)
                return "/".join(url)
    return url

# Run test suite
class run_tests(Command):

    description="Run test suite"


    # I did originally try using 'verbose' as the option but it turns
    # out that is builtin and defaults to 1 (--quiet is also builtin
    # and forces verbose to 0)
    user_options=[
        ("show-tests", "s", "Show each test being run"),
        ]

    # see if you can find boolean_options documented anywhere
    boolean_options=['show-tests']

    def initialize_options(self):
        self.show_tests=0

    def finalize_options(self):
        pass

    def run(self):
        import unittest
        import tests
        tests.setup()
        suite=unittest.TestLoader().loadTestsFromModule(tests)
        # verbosity of zero doesn't print anything, one prints a dot
        # per test and two prints each test name
        result=unittest.TextTestRunner(verbosity=self.show_tests+1).run(suite)
        if not result.wasSuccessful():
            sys.exit(1)

# A hack we dont't document
class build_test_extension(Command):
    description="Compiles APSW test loadable extension"

    user_options=[]

    def initialize_options(self):
        pass
    def finalize_options(self):
        pass
    def run(self):
        # On 64 bit windows we have to use MSVC
        if sys.platform=='win32': # yes even on 64 bit
            try:
                import platform
                if platform.architecture()[0]=='64bit':
                    res=os.system("cl /Gd src/testextension.c /I sqlite3 /I . /DDLL /LD /link /export:sqlite3_extension_init /export:alternate_sqlite3_extension_init /out:testextension.sqlext")
                    if res!=0:
                        raise RuntimeError("Building test extension failed")
                    return
            except ImportError:
                pass
        shared="shared"
        if sys.platform.startswith("darwin"):
            shared="bundle"
        res=os.system("gcc -fPIC -%s -o testextension.sqlext -Isqlite3 -I. src/testextension.c" % (shared,))
        if res!=0:
            raise RuntimeError("Building test extension failed")

# Another hack.  Visual Studio 2008 & 2010 ship with 64
# compilers, headers and the Windows SDK but claims it doesn't and
# distutils can't find it.  The separate Windows SDK can't find this
# and gets very confused not to mention being one of the buggiest cmd
# scripts I have ever seen.  This hack just sets some environment
# variables directly since all the "proper" ways are very broken.
class win64hackvars(Command):
    description="Set env vars for Visual Studio 2008/2010 Express 64 bit"

    user_options=[]
    def initialize_options(self): pass
    def finalize_options(self): pass
    def run(self):
        vcver=9
        if sys.version_info>=(3,3):
            vcver=10
        sdkdir=r"C:\Program Files\Microsoft SDKs\Windows\v6.0A"
        vsdir=r"C:\Program Files (x86)\Microsoft Visual Studio %d.0\VC" % vcver
        assert os.path.isdir(sdkdir), "Expected sdk dir "+sdkdir
        assert os.path.isdir(vsdir), "Expected visual studio dir "+vsdir
        os.environ["PATH"]=r"%s\bin\amd64;%s\bin" % (vsdir, sdkdir)
        os.environ["INCLUDE"]=r"%s\include;%s\include" % (vsdir, sdkdir)
        os.environ["LIB"]=r"%s\lib\amd64;%s\lib\x64" % (vsdir, sdkdir)
        os.environ["DISTUTILS_USE_SDK"]="1"
        os.environ["MSSdk"]=sdkdir

# deal with various python version compatibility issues with how
# to treat returned web data as lines of text
def fixupcode(code):
    if sys.version_info<(2,5):
        if type(code)!=str:
            code=code.read()

    if sys.version_info>=(3,0):
        if type(code)!=bytes:
            code=code.read()
        if type(code)==bytes:
            code=code.decode("iso8859-1")

    if type(code)==str:
        return [l+"\n" for l in code.split("\n")]

    return code

fetch_parts=[]

class fetch(Command):
    description="Automatically downloads SQLite and components"
    user_options=[
        ("version=", None, "Which version of SQLite/components to get (default current)"),
        ("missing-checksum-ok", None, "Continue on a missing checksum (default abort)"),
        ("sqlite", None, "Download SQLite amalgamation"),
        ("all", None, "Download all downloadable components"),
        ]
    fetch_options=['sqlite']
    boolean_options=fetch_options+['all', 'missing-checksum-ok']

    def initialize_options(self):
        self.version=None
        self.sqlite=False
        self.all=False
        self.missing_checksum_ok=False

    def finalize_options(self):
        # If all is selected then turn on all components
        global fetch_parts
        if self.all:
            for i in self.fetch_options:
                    setattr(self, i, True)
        for i in self.fetch_options:
            fetch_parts.append(i)

    def run(self):
        # work out the version
        if self.version is None:
            write("  Getting download page to work out current SQLite version")
            page=self.download("https://sqlite.org/download.html", text=True, checksum=False)
            match=re.search(r'sqlite-amalgamation-3([0-9][0-9])([0-9][0-9])([0-9][0-9])\.zip', page)
            if match:
                self.version="3.%d.%d.%d" % tuple([int(match.group(n)) for n in range(1,4)])
                if self.version.endswith(".0"):
                    self.version=self.version[:-len(".0")]
            else:
                write("Unable to determine current SQLite version.  Use --version=VERSION", sys.stderr)
                write("to set version - eg setup.py fetch --version=3.6.18", sys.stderr)
                sys.exit(17)
            write("    Version is "+self.version)
        # now get each selected component
        downloaded=0

        if not self.version.startswith("fossil"):
            v=[int(x) for x in self.version.split(".")]
            if len(v)<4:
                v.append(0)
            self.webversion="%d%02d%02d%02d" % tuple(v)

        ## The amalgamation
        if self.sqlite:
            if self.version.startswith("fossil"):
                write("  Getting code from fossil")
            else:
                write("  Getting the SQLite amalgamation")

            if self.version.startswith("fossil"):
                if self.version=="fossil":
                    uuid="trunk"
                else:
                    showmsg=False
                    if not self.version.startswith("fossil-"):
                        showmsg=True
                    else:
                        uuid=self.version.split("-", 1)[1]
                        if not uuid:
                            showmsg=True
                    if showmsg:
                        write("Use fossil-HASH to identify a particular commit", sys.stderr)
                        write("eg  fossil-3a82c8e6", sys.stderr)
                        sys.exit(18)

                AURL="https://sqlite.org/src/zip/sqlite3.zip?uuid="+uuid
                checksum=False
            else:
                if sys.platform=="win32":
                    AURL="https://sqlite.org/sqlite-amalgamation-%s.zip" % (self.webversion,)
                else:
                    AURL="https://sqlite.org/sqlite-autoconf-%s.tar.gz" % (self.webversion,)
                checksum=True

            AURL=fixup_download_url(AURL)

            data=self.download(AURL, checksum=checksum)

            if AURL.endswith(".zip"):
                zip=zipfile.ZipFile(data, "r")
                for name in "sqlite3.c", "sqlite3.h", "sqlite3ext.h":
                    write("Extracting", name)
                    f=[n for n in zip.namelist() if n.endswith(name)]
                    if len(f)!=1:
                        raise Exception("Can't find %s in zip.  Candidates are %s" % (name, f))
                    # Work around SQLite 3.7.13 bug where a symbol was
                    # declared SQLITE_API and extern
                    data=zip.read(f[0])
                    if name=="sqlite3.c":
                        data=data.decode("utf8")
                        data=data.replace("SQLITE_API extern", "SQLITE_API")
                        data=data.encode("utf8")
                    open(name, "wb").write(data)
                zip.close()
            else:
                # we need to run configure to get various -DHAVE_foo flags on non-windows platforms
                # delete existing sqlite3 directory if it exists, but save sqlite3config.h if it exists
                sqlite3config_h=None
                if os.path.exists("sqlite3/sqlite3config.h"):
                    sqlite3config_h=read_whole_file("sqlite3/sqlite3config.h", "rb")
                if os.path.exists('sqlite3'):
                    for dirpath, dirnames, filenames in os.walk('sqlite3', topdown=False):
                        for file in filenames:
                            os.remove(os.path.join(dirpath, file))
                        for dir in dirnames:
                            os.rmdir(os.path.join(dirpath, dir))
                    os.rmdir('sqlite3')
                if self.version.startswith("fossil"):
                    zip=zipfile.ZipFile(data, "r")
                    for name in zip.namelist():
                        # extract
                        if name.endswith("/"):
                            os.mkdir(name)
                        else:
                            open(name, "wb").write(zip.read(name))
                    zip.close()
                else:
                    # if you get an exception here it is likely that you don't have the python zlib module
                    import zlib
                    tar=tarfile.open("nonexistentname to keep old python happy", 'r', data)
                    configmember=None
                    for member in tar.getmembers():
                        tar.extract(member)
                        # find first file named configure
                        if not configmember and member.name.endswith("/configure"):
                            configmember=member
                    tar.close()
                    # the directory name has changed a bit with each release so try to work out what it is
                    if not configmember:
                        write("Unable to determine directory it extracted to.", dest=sys.stderr)
                        sys.exit(19)
                    dirname=configmember.name.split('/')[0]
                    os.rename(dirname, 'sqlite3')
                os.chdir('sqlite3')
                if self.version.startswith("fossil"):
                    write("    Building amalgamation from fossil")
                    res=os.system("make TOP=. -f Makefile.linux-gcc sqlite3.c && cp src/sqlite3ext.h .")
                    defs=[]
                    if sqlite3config_h:
                        open("sqlite3config.h", "wb").write(sqlite3config_h)
                else:
                    write("    Running configure to work out SQLite compilation flags")
                    res=os.system("./configure >/dev/null")
                    defline=None
                    for line in read_whole_file("Makefile", "rtU").split("\n"):
                        if line.startswith("DEFS = "):
                            defline=line
                            break
                    if not defline:
                        write("Unable to determine compile flags.  Create sqlite3/sqlite3config.h to manually set.", sys.stderr)
                        sys.exit(18)
                    defs=[]
                    for part in shlex.split(defline):
                        if part.startswith("-DHAVE"):
                            part=part[2:]
                            if '=' in part:
                                part=part.split('=', 1)
                            else:
                                part=(part, )
                            defs.append(part)
                if res!=0:
                    raise ValueError("Command execution failed")
                if defs:
                    op=open("sqlite3config.h", "wt")
                    op.write("""
/* This file was generated by parsing how configure altered the Makefile
   which isn't used when building python extensions.  It is specific to the
   machine and developer components on which it was run. */
   \n""")

                    for define in defs:
                        op.write('#define %s %s\n' % tuple(define))
                    op.close()
                os.chdir("..")
            downloaded+=1

        if not downloaded:
            write("You didn't specify any components to fetch.  Use")
            write("   setup.py fetch --help")
            write("for a list and details")
            raise ValueError("No components downloaded")

    # A function for verifying downloads
    def verifyurl(self, url, data):
        d=["%s" % (len(data),)]
        try:
            import hashlib
            d.append(hashlib.sha1(data).hexdigest())
            d.append(hashlib.md5(data).hexdigest())
        except ImportError:
            import sha
            d.append(sha.new(data).hexdigest())
            import md5
            d.append(md5.new(data).hexdigest())

        write("    Length:", d[0], " SHA1:", d[1], " MD5:", d[2])
        sums=os.path.join(os.path.dirname(__file__), "checksums")
        for line in read_whole_file(sums, "rt").split("\n"):
            line=line.strip()
            if len(line)==0 or line[0]=="#":
                continue
            l=[l.strip() for l in line.split()]
            if len(l)!=4:
                write("Invalid line in checksums file:", line, sys.stderr)
                raise ValueError("Bad checksums file")
            if l[0]==url:
                if l[1:]==d:
                    write("    Checksums verified")
                    return
                if l[1]!=d[0]:
                    write("Length does not match.  Expected", l[1], "download was", d[0])
                if l[2]!=d[1]:
                    write("SHA does not match.  Expected", l[2], "download was", d[1])
                if l[3]!=d[2]:
                    write("MD5 does not match.  Expected", l[3], "download was", d[2])
                write("The download does not match the checksums distributed with APSW.\n"
                      "The download should not have changed since the checksums were\n"
                      "generated.  The cause could be anything from network corruption\n"
                      "to a malicious attack.")
                raise ValueError("Checksums do not match")
        # no matching line
        write("    (Not verified.  No match in checksums file)")
        if not self.missing_checksum_ok:
            raise ValueError("No checksum available.  Use --missing-checksum-ok option to continue")

    # download a url
    def download(self, url, text=False, checksum=True):
        if py3:
            import urllib.request
            urlopen=urllib.request.urlopen
            import io
            bytesio=io.BytesIO
        else:
            import urllib2
            urlopen=urllib2.urlopen
            import cStringIO
            bytesio=cStringIO.StringIO

        write("    Fetching", url)
        count=0
        while True:
            try:
                if count:
                    write("        Try #",str(count+1))
                try:
                    page=urlopen(url).read()
                except:
                    # Degrade to http if https is not supported
                    e=sys.exc_info()[1]
                    if "eof occurred in violation of protocol" in str(e).lower() or getattr(e, "reason")=="unknown url type: https":
                        write("        [Python has https issues - using http instead]")
                        page=urlopen(url.replace("https://", "http://")).read()
                    else:
                        raise
                break
            except:
                write("       Error ", str(sys.exc_info()[1]))
                time.sleep(1.3)
                count+=1
                if count>=5:
                    raise

        if text:
            if py3:
                page=page.decode("iso8859_1")

        if checksum:
            self.verifyurl(url, page)

        if not text:
            page=bytesio(page)

        return page

# We allow enable/omit to be specified to build and then pass them to build_ext
build_enable=None
build_omit=False # PLASTICITY
build_enable_all_extensions=True # PLASTICITY

bparent=build.build
class apsw_build(bparent):
    user_options=bparent.user_options+\
                  [ ("enable=", None, "Enable SQLite options (comma separated list)"),
                    ("omit=", None, "Omit SQLite functionality (comma separated list)"),
                    ("enable-all-extensions", None, "Enable all SQLite extensions"),
                    ]
    boolean_options=bparent.boolean_options+["enable-all-extensions"]

    def initialize_options(self):
        v=bparent.initialize_options(self)
        self.enable=None
        self.omit=None
        self.enable_all_extensions=build_enable_all_extensions
        return v

    def finalize_options(self):
        global build_enable, build_omit, build_enable_all_extensions
        build_enable=self.enable
        build_omit=self.omit
        build_enable_all_extensions=self.enable_all_extensions
        return bparent.finalize_options(self)

def findamalgamation():
    amalgamation=(
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "sqlite3.c"),
        os.path.join(os.path.dirname(os.path.abspath(__file__)), "sqlite3", "sqlite3.c")
        )
    for path in amalgamation:
        if os.path.exists(path):
            return path
    return None

def find_in_path(name):
    for loc in os.getenv("PATH").split(os.pathsep):
        f=os.path.abspath(os.path.join(loc, name))
        if os.path.exists(f) or os.path.exists(f.lower()) or os.path.exists(f.lower()+".exe"):
            return f
    return None

beparent=build_ext.build_ext
class apsw_build_ext(beparent):

    user_options=beparent.user_options+\
                  [ ("enable=", None, "Enable SQLite options (comma separated list)"),
                    ("omit=", None, "Omit SQLite functionality (comma separated list)"),
                    ("enable-all-extensions", None, "Enable all SQLite extensions"),
                    ]
    boolean_options=beparent.boolean_options+["enable-all-extensions"]


    def initialize_options(self):
        v=beparent.initialize_options(self)
        self.enable=build_enable
        self.omit=build_omit
        self.enable_all_extensions=build_enable_all_extensions
        return v

    def finalize_options(self):
        v=beparent.finalize_options(self)

        if self.enable_all_extensions:
            exts=["fts4", "fts3", "fts3_parenthesis", "rtree", "stat4", "json1", "fts5", "rbu"]
            if find_in_path("icu-config"):
                exts.append("icu")
            if not self.enable:
                self.enable=",".join(exts)
            else:
                self.enable=self.enable+","+",".join(exts)

        ext=self.extensions[0]

        if not ext.define_macros: ext.define_macros=[]
        if not ext.depends: ext.depends=[]
        if not ext.include_dirs: ext.include_dirs=[]
        if not ext.library_dirs: ext.library_dirs=[]
        if not ext.libraries: ext.libraries=[]

        # Fixup debug setting
        if self.debug:
            # distutils forces NDEBUG even with --debug so overcome that
            ext.define_macros.append( ('APSW_NO_NDEBUG', '1') )    # double negatives are bad
            ext.define_macros.append( ('APSW_TESTFIXTURES', '1') ) # extra test harness code
            ext.define_macros.append( ('SQLITE_DEBUG', '1') )      # also does NDEBUG mangling
        else:
            ext.define_macros.append( ('NDEBUG', '1') )

        # fork checker?
        if hasattr(os, "fork"):
            ext.define_macros.append( ('APSW_FORK_CHECKER', '1') )
        
        ext.define_macros.append( ("SQLITE_MAX_COLUMN", "32767") ) # PLASTICITY
        ext.define_macros.append( ("SQLITE_MAX_VARIABLE_NUMBER", "99999") ) # PLASTICITY

        # SQLite 3
        # Look for amalgamation in our directory or in sqlite3 subdirectory

        path=findamalgamation()
        if path:
            if sys.platform=="win32":
                # double quotes get consumed by windows arg processing
                ext.define_macros.append( ('APSW_USE_SQLITE_AMALGAMATION', '\\"'+path+'\\"') )
            else:
                ext.define_macros.append( ('APSW_USE_SQLITE_AMALGAMATION', '"'+path+'"') )
            ext.depends.append(path)
            # we also add the directory to include path since icu tries to use it
            ext.include_dirs.append(os.path.dirname(path))
            write("SQLite: Using amalgamation", path)
            load_extension=True
        else:
            load_extension=False
            d=os.path.join(os.path.dirname(os.path.abspath(__file__)), "sqlite3")
            if os.path.isdir(d):
                write("SQLite: Using include/libraries in sqlite3 subdirectory")
                ext.include_dirs.append(d)
                ext.library_dirs.append(d)
            else:
                write("SQLite: Using system sqlite include/libraries")
            ext.libraries.append('sqlite3')

        s3config=os.path.join(os.path.dirname(os.path.abspath(__file__)), "sqlite3", "sqlite3config.h")
        if os.path.exists(s3config):
            if sys.platform=="win32":
               ext.define_macros.append( ('APSW_USE_SQLITE_CONFIG', '\\"'+s3config+'\\"') )
            else:
                ext.define_macros.append( ('APSW_USE_SQLITE_CONFIG', '"'+s3config+'"') )

        # enables
        addicuinclib=False
        if self.enable:
            for e in self.enable.split(","):
                e=e.strip()
                if e.lower()=="load_extension":
                    load_extension=True
                    continue
                ext.define_macros.append( ("SQLITE_ENABLE_"+e.upper(), 1) )
                if e.upper()=="ICU":
                    addicuinclib=True
                os.putenv("APSW_TEST_"+e.upper(), "1")
                # See issue #55 where I had left off the 3 in fts3.  This code
                # tries to catch misspelling the name of an extension.
                # However the SQLITE_ENABLE prefix is also used by other
                # options - see https://sqlite.org/compile.html but almost
                # all of those have _ in them, so our abbreviated and
                # hopefully future proof test
                if "_" not in e.lower() and \
                       "memsys" not in e.lower() and \
                       e.lower() not in ("fts4", "fts3", "rtree", "icu", "iotrace",
                                         "stat2", "stat3", "stat4", "dbstat_vtab",
                                         "fts5", "json1", "rbu"):
                    write("Unknown enable "+e, sys.stderr)
                    raise ValueError("Bad enable "+e)

        # omits
        if self.omit:
            for e in self.omit.split(","):
                e=e.strip()
                if e.lower()=="load_extension":
                    load_extension=False
                ext.define_macros.append( ("SQLITE_OMIT_"+e.upper(), 1) )

        if not load_extension:
            ext.define_macros.append( ("SQLITE_OMIT_LOAD_EXTENSION", 1) )

        # icu
        if addicuinclib:
            foundicu=False
            kwargs={}
            if sys.version_info>=(2, 6):
                # if posix is true then quotes get stripped such as from -Dfoo="bar"
                kwargs["posix"]=False
            for part in shlex.split(os.popen("icu-config --cppflags", "r").read(), **kwargs):
                if part.startswith("-I"):
                    ext.include_dirs.append(part[2:])
                    foundicu=True
                elif part.startswith("-D"):
                    part=part[2:]
                    if '=' in part:
                        part=tuple(part.split('=', 1))
                    else:
                        part=(part, '1')
                    ext.define_macros.append(part)
                    foundicu=True

            for part in shlex.split(os.popen("icu-config --ldflags", "r").read(), **kwargs):
                if part.startswith("-L"):
                    ext.library_dirs.append(part[2:])
                    foundicu=True
                elif part.startswith("-l"):
                    ext.libraries.append(part[2:])
                    foundicu=True

            if foundicu:
                write("ICU: Added includes, flags and libraries from icu-config")
            else:
                write("ICU: Unable to determine includes/libraries for ICU using icu-config")
                write("ICU: You will need to manually edit setup.py or setup.cfg to set them")

        # shell
        if not os.path.exists("src/shell.c") or \
               os.path.getmtime("src/shell.c")<os.path.getmtime("tools/shell.py") or \
               os.path.getmtime(__file__)>os.path.getmtime("src/shell.c"):
            create_c_file("tools/shell.py", "src/shell.c")

        # done ...
        return v

    def run(self):
        v=beparent.run(self)
        return v

sparent=sdist.sdist
class apsw_sdist(sparent):

    user_options=sparent.user_options+[
        ("add-doc", None, "Includes built documentation from doc/build/html into source"),
        ]

    boolean_options=sparent.boolean_options+["add-doc"]

    def initialize_options(self):
        sparent.initialize_options(self)
        self.add_doc=False
        # Were we made from a source archive?  If so include the help again
        if os.path.isfile("doc/index.html") and os.path.isfile("doc/_sources/pysqlite.txt"):
            self.add_doc=True
        self.use_defaults=False # they are useless

        # Make sure the manifest is regenerated
        self.force_manifest=True

        # Now do some chicanery.  If a source distribution is requested and
        # fetch --sqlite was requested then make sure the sqlite amalgamation
        # ends up as part of the source distribution.
        if fetch_parts:
            # Use a temporary file for the manifest
            tmpmanifest="MANIFEST.in.tmp"
            self.template=tmpmanifest
            try:
                os.remove(tmpmanifest)
            except:
                pass
            min=open("MANIFEST.in", "rU")
            mout=open(tmpmanifest, "wt")
            for line in min:
                mout.write(line)
            min.close()
            # os.path.relpath emulation
            if "sqlite" in fetch_parts:
                amalgamationpath=findamalgamation()
                amalrelpath=amalgamationpath[len(os.path.dirname(os.path.abspath(__file__)))+1:]
                mout.write("include "+amalrelpath+"\n")
                # also include headers and extension headers
                mout.write("include "+amalrelpath.replace("sqlite3.c", "sqlite3.h")+"\n")
                mout.write("include "+amalrelpath.replace("sqlite3.c", "sqlite3ext.h")+"\n")
                if os.path.exists("sqlite3/sqlite3config.h"):
                    mout.write("include sqlite3/sqlite3config.h\n")

            mout.close()

    def run(self):
        v=sparent.run(self)
        if self.add_doc:
            if len(list(help_walker('')))==0:
                raise Exception("The help is not built")
            for archive in self.get_archive_files():
                add_doc(archive, self.distribution.get_fullname())
        return v

def help_walker(arcdir):
    # Provides a list of (archive name, disk name) for all the help files
    if os.path.isfile("doc/index.html") and os.path.isfile("doc/_sources/pysqlite.txt"):
        topdir="doc/"
    else:
        topdir="doc/build/html/"
    for dirpath, _, filenames in os.walk(topdir):
        prefix=dirpath[len(topdir):]
        for f in filenames:
            yield os.path.join(arcdir, "doc", prefix, f), os.path.join(dirpath, f)

def add_doc(archive, topdir):
    write("Add help files to",archive)
    if archive.endswith(".tar") or ".tar." in archive:
        if archive.endswith(".Z"):
            raise Exception("tarfile module doesn't support old school compress so we can't add doc "+archive)
        fmt=""
        if archive.endswith(".gz") or archive.endswith(".tgz"):
            fmt=":gz"
        elif archive.endswith(".bz2") or archive.endswith(".tbz2"):
            fmt=":bz2"
        oldarchive=tarfile.open(archive)
        newarchive=tarfile.open(archive+"-", mode="w"+fmt)
        for mem in oldarchive.getmembers():
            newarchive.addfile(mem, oldarchive.extractfile(mem))
        oldarchive.close()
        for arcname, fname in help_walker(topdir):
            newarchive.add(fname, arcname)
        newarchive.close()
        os.rename(archive+"-", archive)
    elif archive.endswith(".zip"):
        ofile=zipfile.ZipFile(archive, "a", zipfile.ZIP_DEFLATED)
        for arcname, fname in help_walker(topdir):
            ofile.write(fname, arcname)
        ofile.close()
    else:
        raise Exception("Don't know what to do with "+archive)

def create_c_file(src, dest):
    # Transforms Python src into C dest as a sequence of strings.
    # Because of the pathetic microsoft compiler we have to break it
    # up into small chunks
    out=["/* Automatically generated by setup.py from "+src+" */", ""]
    percents=1
    size=0
    for line in read_whole_file(src, "rt").split("\n"):
        if "if__name__=='__main__':" in line.replace(" ",""):
            break
        if line.strip().startswith('#'): # full line comment
            continue
        if line.strip()=="import apsw":
            continue
        size=size+len(line)
        comma=size>32000
        if comma:
            size=0
            percents+=1
        line=line.replace("\\", "\\\\").\
              replace('"', '\\"')
        out.append('  "'+line.rstrip()+'\\n"')
        if comma:
            out[-1]=out[-1]+","
    if out[-1].endswith(","):
        out[-1]=out[-1][:-1]
    out[1]='"%s",' % ("%s" * percents,)
    write_whole_file(dest, "wt", "\n".join(out))

# We depend on every .[ch] file in src
depends=[f for f in glob.glob("src/*.[ch]") if f!="src/apsw.c"]
for f in (findamalgamation(), ):
    if f:
        depends.append(f)
# we produce a .c file from this
depends.append("tools/shell.py")

# work out version number
version=read_whole_file(os.path.join("src", "apswversion.h"), "rt").split()[2].strip('"')

# msi can't use normal version numbers because distutils is retarded,
# so mangle ours to suit it
if "bdist_msi" in sys.argv:
    if version.endswith("-r1"):
        version=version[:-len("-r1")]
    else:
        assert False, "MSI version needs help"
    version=[int(v) for v in re.split(r"[^\d]+", version)]
    # easy pad to 3 items long
    while len(version)<3:
        version.append(0)
    # 4 is our normal length (eg 3.7.3-r1) but sometimes it is more eg
    # 3.7.16.1-r1 so combine last elements if longer than 4
    while len(version)>4:
        version[-2]=10*version[-2]+version-1
        del version[-1]
    # combine first two elements
    if len(version)>3:
        version[0]=100*version[0]+version[1]
        del version[1]

    version=".".join([str(v) for v in version])

setup(name="apsw",
      version=version,
      description="Another Python SQLite Wrapper",
      long_description=\
"""A Python wrapper for the SQLite embedded relational database engine.
In contrast to other wrappers such as pysqlite it focuses on being
a minimal layer over SQLite attempting just to translate the
complete SQLite API into Python.""",
      author="Roger Binns",
      author_email="rogerb@rogerbinns.com",
      url="https://github.com/rogerbinns/apsw/",
      classifiers=[
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "License :: OSI Approved",
    "Operating System :: OS Independent",
    "Programming Language :: C",
    "Programming Language :: Python :: 2",
    "Programming Language :: Python :: 3",
    "Topic :: Database :: Front-Ends",
    ],
      keywords=["database", "sqlite"],
      license="OSI Approved ::",

      ext_modules=[Extension("apsw",
                             ["src/apsw.c", "src/sqlite3.c"], # PLASTICITY
                             include_dirs=include_dirs,
                             library_dirs=library_dirs,
                             libraries=libraries,
                             define_macros=define_macros,
                             depends=depends)],


      cmdclass={'test': run_tests,
                'build_test_extension': build_test_extension,
                'fetch': fetch,
                'build_ext': apsw_build_ext,
                'build': apsw_build,
                'sdist': apsw_sdist,
                'win64hackvars': win64hackvars}
      )





from __future__ import division
from __future__ import absolute_import
#typing
from collections import Counter

#overrides
from sklearn.utils.linear_assignment_ import linear_assignment
import numpy as np
import torch

from allennlp.training.metrics.metric import Metric

class ConllCorefScores(Metric):
    def __init__(self)        :
        self.scorers = [Scorer(m) for m in (Scorer.muc, Scorer.b_cubed, Scorer.ceafe)]

    #overrides
    def __call__(self,  # type: ignore
                 top_spans              ,
                 antecedent_indices              ,
                 predicted_antecedents              ,
                 metadata_list                      ):
        u"""
        Parameters
        ----------
        top_spans : ``torch.Tensor``
            (start, end) indices for all spans kept after span pruning in the model.
            Expected shape: (batch_size, num_spans, 2)
        antecedent_indices : ``torch.Tensor``
            For each span, the indices of all allowed antecedents for that span.  This is
            independent of the batch dimension, as it's just based on order in the document.
            Expected shape: (num_spans, num_antecedents)
        predicted_antecedents: ``torch.Tensor``
            For each span, this contains the index (into antecedent_indices) of the most likely
            antecedent for that span.
            Expected shape: (batch_size, num_spans)
        metadata_list : ``List[Dict[str, Any]]``
            A metadata dictionary for each instance in the batch.  We use the "clusters" key from
            this dictionary, which has the annotated gold coreference clusters for that instance.
        """
        top_spans, antecedent_indices, predicted_antecedents = self.unwrap_to_tensors(top_spans,
                                                                                      antecedent_indices,
                                                                                      predicted_antecedents)
        for i, metadata in enumerate(metadata_list):
            gold_clusters, mention_to_gold = self.get_gold_clusters(metadata[u"clusters"])
            predicted_clusters, mention_to_predicted = self.get_predicted_clusters(top_spans[i],
                                                                                   antecedent_indices,
                                                                                   predicted_antecedents[i])
            for scorer in self.scorers:
                scorer.update(predicted_clusters, gold_clusters, mention_to_predicted, mention_to_gold)

    #overrides
    def get_metric(self, reset       = False)                              :
        metrics = (lambda e: e.get_precision(), lambda e: e.get_recall(), lambda e: e.get_f1())
        precision, recall, f1_score = tuple(sum(metric(e) for e in self.scorers) / len(self.scorers)
                                            for metric in metrics)
        if reset:
            self.reset()
        return precision, recall, f1_score

    #overrides
    def reset(self):
        self.scorers = [Scorer(metric) for metric in (Scorer.muc, Scorer.b_cubed, Scorer.ceafe)]

    @staticmethod
    def get_gold_clusters(gold_clusters):
        gold_clusters = [tuple(tuple(m) for m in gc) for gc in gold_clusters]
        mention_to_gold = {}
        for gold_cluster in gold_clusters:
            for mention in gold_cluster:
                mention_to_gold[mention] = gold_cluster
        return gold_clusters, mention_to_gold

    @staticmethod
    def get_predicted_clusters(top_spans              ,
                               antecedent_indices              ,
                               predicted_antecedents              ):                                            
                                                                                                  
                                                                                                               
        # Pytorch 0.4 introduced scalar tensors, so our calls to tuple() and such below don't
        # actually give ints unless we convert to numpy first.  So we do that here.
        top_spans = top_spans.numpy()  # (num_spans, 2)
        antecedent_indices = antecedent_indices.numpy()  # (num_spans, num_antecedents)
        predicted_antecedents = predicted_antecedents.numpy()  # (num_spans,)

        predicted_clusters_to_ids                             = {}
        clusters                              = []
        for i, predicted_antecedent in enumerate(predicted_antecedents):
            if predicted_antecedent < 0:
                continue

            # Find predicted index in the antecedent spans.
            predicted_index = antecedent_indices[i, predicted_antecedent]
            # Must be a previous span.
            assert i > predicted_index
            antecedent_span                  = tuple(top_spans[predicted_index])  # type: ignore

            # Check if we've seen the span before.
            if antecedent_span in list(predicted_clusters_to_ids.keys()):
                predicted_cluster_id      = predicted_clusters_to_ids[antecedent_span]
            else:
                # We start a new cluster.
                predicted_cluster_id = len(clusters)
                clusters.append([antecedent_span])
                predicted_clusters_to_ids[antecedent_span] = predicted_cluster_id

            mention                  = tuple(top_spans[i])  # type: ignore
            clusters[predicted_cluster_id].append(mention)
            predicted_clusters_to_ids[mention] = predicted_cluster_id

        # finalise the spans and clusters.
        final_clusters = [tuple(cluster) for cluster in clusters]
        # Return a mapping of each mention to the cluster containing it.
        mention_to_cluster                                                     = dict((
                mention, final_clusters[cluster_id])
                for mention, cluster_id in list(predicted_clusters_to_ids.items()))

        return final_clusters, mention_to_cluster


ConllCorefScores = Metric.register(u"conll_coref_scores")(ConllCorefScores)

class Scorer(object):
    u"""
    Mostly borrowed from <https://github.com/clarkkev/deep-coref/blob/master/evaluation.py>
    """
    def __init__(self, metric):
        self.precision_numerator = 0
        self.precision_denominator = 0
        self.recall_numerator = 0
        self.recall_denominator = 0
        self.metric = metric

    def update(self, predicted, gold, mention_to_predicted, mention_to_gold):
        if self.metric == self.ceafe:
            p_num, p_den, r_num, r_den = self.metric(predicted, gold)
        else:
            p_num, p_den = self.metric(predicted, mention_to_gold)
            r_num, r_den = self.metric(gold, mention_to_predicted)
        self.precision_numerator += p_num
        self.precision_denominator += p_den
        self.recall_numerator += r_num
        self.recall_denominator += r_den

    def get_f1(self):
        precision = 0 if self.precision_denominator == 0 else\
            self.precision_numerator / float(self.precision_denominator)
        recall = 0 if self.recall_denominator == 0 else\
            self.recall_numerator / float(self.recall_denominator)
        return 0 if precision + recall == 0 else 2 * precision * recall / (precision + recall)

    def get_recall(self):
        if self.recall_numerator == 0:
            return 0
        else:
            return self.recall_numerator / float(self.recall_denominator)

    def get_precision(self):
        if self.precision_numerator == 0:
            return 0
        else:
            return self.precision_numerator / float(self.precision_denominator)

    def get_prf(self):
        return self.get_precision(), self.get_recall(), self.get_f1()

    @staticmethod
    def b_cubed(clusters, mention_to_gold):
        u"""
        Averaged per-mention precision and recall.
        <https://pdfs.semanticscholar.org/cfe3/c24695f1c14b78a5b8e95bcbd1c666140fd1.pdf>
        """
        numerator, denominator = 0, 0
        for cluster in clusters:
            if len(cluster) == 1:
                continue
            gold_counts = Counter()
            correct = 0
            for mention in cluster:
                if mention in mention_to_gold:
                    gold_counts[tuple(mention_to_gold[mention])] += 1
            for cluster2, count in list(gold_counts.items()):
                if len(cluster2) != 1:
                    correct += count * count
            numerator += correct / float(len(cluster))
            denominator += len(cluster)
        return numerator, denominator

    @staticmethod
    def muc(clusters, mention_to_gold):
        u"""
        Counts the mentions in each predicted cluster which need to be re-allocated in
        order for each predicted cluster to be contained by the respective gold cluster.
        <http://aclweb.org/anthology/M/M95/M95-1005.pdf>
        """
        true_p, all_p = 0, 0
        for cluster in clusters:
            all_p += len(cluster) - 1
            true_p += len(cluster)
            linked = set()
            for mention in cluster:
                if mention in mention_to_gold:
                    linked.add(mention_to_gold[mention])
                else:
                    true_p -= 1
            true_p -= len(linked)
        return true_p, all_p

    @staticmethod
    def phi4(gold_clustering, predicted_clustering):
        u"""
        Subroutine for ceafe. Computes the mention F measure between gold and
        predicted mentions in a cluster.
        """
        return 2 * len([mention for mention in gold_clustering if mention in predicted_clustering])\
               / float(len(gold_clustering) + len(predicted_clustering))

    @staticmethod
    def ceafe(clusters, gold_clusters):
        u"""
        Computes the  Constrained EntityAlignment F-Measure (CEAF) for evaluating coreference.
        Gold and predicted mentions are aligned into clusterings which maximise a metric - in
        this case, the F measure between gold and predicted clusters.

        <https://www.semanticscholar.org/paper/On-Coreference-Resolution-Performance-Metrics-Luo/de133c1f22d0dfe12539e25dda70f28672459b99>
        """
        clusters = [cluster for cluster in clusters if len(cluster) != 1]
        scores = np.zeros((len(gold_clusters), len(clusters)))
        for i, gold_cluster in enumerate(gold_clusters):
            for j, cluster in enumerate(clusters):
                scores[i, j] = Scorer.phi4(gold_cluster, cluster)
        matching = linear_assignment(-scores)
        similarity = sum(scores[matching[:, 0], matching[:, 1]])
        return similarity, len(clusters), similarity, len(gold_clusters)



# pylint: disable=no-self-use,invalid-name


from __future__ import division
from __future__ import absolute_import
import os
import pytest

from allennlp.common.testing import AllenNlpTestCase
from allennlp.models.archival import load_archive
from allennlp.predictors import Predictor
from allennlp.predictors.wikitables_parser import (SEMPRE_ABBREVIATIONS_PATH, SEMPRE_GRAMMAR_PATH)

class TestWikiTablesParserPredictor(AllenNlpTestCase):
    def setUp(self):
        super(TestWikiTablesParserPredictor, self).setUp()
        self.should_remove_sempre_abbreviations = not os.path.exists(SEMPRE_ABBREVIATIONS_PATH)
        self.should_remove_sempre_grammar = not os.path.exists(SEMPRE_GRAMMAR_PATH)

    def tearDown(self):
        super(TestWikiTablesParserPredictor, self).tearDown()
        if self.should_remove_sempre_abbreviations and os.path.exists(SEMPRE_ABBREVIATIONS_PATH):
            os.remove(SEMPRE_ABBREVIATIONS_PATH)
        if self.should_remove_sempre_grammar and os.path.exists(SEMPRE_GRAMMAR_PATH):
            os.remove(SEMPRE_GRAMMAR_PATH)

    def test_uses_named_inputs(self):
        inputs = {
                u"question": u"names",
                u"table": u"name\tdate\nmatt\t2017\npradeep\t2018"
        }

        archive_path = self.FIXTURES_ROOT / u'semantic_parsing' / u'wikitables' / u'serialization' / u'model.tar.gz'
        archive = load_archive(archive_path)
        predictor = Predictor.from_archive(archive, u'wikitables-parser')

        result = predictor.predict_json(inputs)

        action_sequence = result.get(u"best_action_sequence")
        if action_sequence:
            # We don't currently disallow endless loops in the decoder, and an untrained seq2seq
            # model will easily get itself into a loop.  An endless loop isn't a finished logical
            # form, so decoding doesn't return any finished states, which means no actions.  So,
            # sadly, we don't have a great test here.  This is just testing that the predictor
            # runs, basically.
            assert len(action_sequence) > 1
            assert all([isinstance(action, unicode) for action in action_sequence])

            logical_form = result.get(u"logical_form")
            assert logical_form is not None

    def test_answer_present(self):
        inputs = {
                u"question": u"Who is 18 years old?",
                u"table": u"Name\tAge\nShallan\t16\nKaladin\t18"
        }

        archive_path = self.FIXTURES_ROOT / u'semantic_parsing' / u'wikitables' / u'serialization' / u'model.tar.gz'
        archive = load_archive(archive_path)
        predictor = Predictor.from_archive(archive, u'wikitables-parser')

        result = predictor.predict_json(inputs)
        answer = result.get(u"answer")
        assert answer is not None

    def test_answer_present_with_batch_predict(self):
        inputs = [{
                u"question": u"Who is 18 years old?",
                u"table": u"Name\tAge\nShallan\t16\nKaladin\t18"
        }]

        archive_path = self.FIXTURES_ROOT / u'semantic_parsing' / u'wikitables' / u'serialization' / u'model.tar.gz'
        archive = load_archive(archive_path)
        predictor = Predictor.from_archive(archive, u'wikitables-parser')

        result = predictor.predict_batch_json(inputs)
        answer = result[0].get(u"answer")
        assert answer is not None

TestWikiTablesParserPredictor = pytest.mark.java(TestWikiTablesParserPredictor)



# pylint: disable=no-self-use,invalid-name


from __future__ import with_statement
from __future__ import absolute_import
from numpy.testing import assert_almost_equal
import pytest
import torch

from allennlp.common import Params
from allennlp.common.checks import ConfigurationError
from allennlp.modules import FeedForward
from allennlp.nn import InitializerApplicator
from allennlp.common.testing import AllenNlpTestCase


class TestFeedForward(AllenNlpTestCase):
    def test_init_checks_hidden_dim_consistency(self):
        with pytest.raises(ConfigurationError):
            FeedForward(2, 4, [5, 5], u'relu')

    def test_init_checks_activation_consistency(self):
        with pytest.raises(ConfigurationError):
            FeedForward(2, 4, 5, [u'relu', u'relu'])

    def test_forward_gives_correct_output(self):
        params = Params({
                u'input_dim': 2,
                u'hidden_dims': 3,
                u'activations': u'relu',
                u'num_layers': 2
                })
        feedforward = FeedForward.from_params(params)

        constant_init = lambda tensor: torch.nn.init.constant_(tensor, 1.)
        initializer = InitializerApplicator([(u".*", constant_init)])
        initializer(feedforward)

        input_tensor = torch.FloatTensor([[-3, 1]])
        output = feedforward(input_tensor).data.numpy()
        assert output.shape == (1, 3)
        # This output was checked by hand - ReLU makes output after first hidden layer [0, 0, 0],
        # which then gets a bias added in the second layer to be [1, 1, 1].
        assert_almost_equal(output, [[1, 1, 1]])



# pylint: disable=no-self-use,invalid-name


from __future__ import absolute_import
from allennlp.common.testing import AllenNlpTestCase
from allennlp.data.tokenizers.token import Token
from allennlp.data.tokenizers.word_splitter import LettersDigitsWordSplitter
from allennlp.data.tokenizers.word_splitter import SimpleWordSplitter
from allennlp.data.tokenizers.word_splitter import SpacyWordSplitter
try:
    from itertools import izip
except:
    izip = zip



class TestSimpleWordSplitter(AllenNlpTestCase):
    def setUp(self):
        super(TestSimpleWordSplitter, self).setUp()
        self.word_splitter = SimpleWordSplitter()

    def test_tokenize_handles_complex_punctuation(self):
        sentence = u"this (sentence) has 'crazy' \"punctuation\"."
        expected_tokens = [u"this", u"(", u"sentence", u")", u"has", u"'", u"crazy", u"'", u'"',
                           u"punctuation", u'"', u"."]
        tokens = [t.text for t in self.word_splitter.split_words(sentence)]
        assert tokens == expected_tokens

    def test_tokenize_handles_contraction(self):
        sentence = u"it ain't joe's problem; would've been yesterday"
        expected_tokens = [u"it", u"ai", u"n't", u"joe", u"'s", u"problem", u";", u"would", u"'ve", u"been",
                           u"yesterday"]
        tokens = [t.text for t in self.word_splitter.split_words(sentence)]
        assert tokens == expected_tokens

    def test_batch_tokenization(self):
        sentences = [u"This is a sentence",
                     u"This isn't a sentence.",
                     u"This is the 3rd sentence."
                     u"Here's the 'fourth' sentence."]
        batch_split = self.word_splitter.batch_split_words(sentences)
        separately_split = [self.word_splitter.split_words(sentence) for sentence in sentences]
        assert len(batch_split) == len(separately_split)
        for batch_sentence, separate_sentence in izip(batch_split, separately_split):
            assert len(batch_sentence) == len(separate_sentence)
            for batch_word, separate_word in izip(batch_sentence, separate_sentence):
                assert batch_word.text == separate_word.text

    def test_tokenize_handles_multiple_contraction(self):
        sentence = u"wouldn't've"
        expected_tokens = [u"would", u"n't", u"'ve"]
        tokens = [t.text for t in self.word_splitter.split_words(sentence)]
        assert tokens == expected_tokens

    def test_tokenize_handles_final_apostrophe(self):
        sentence = u"the jones' house"
        expected_tokens = [u"the", u"jones", u"'", u"house"]
        tokens = [t.text for t in self.word_splitter.split_words(sentence)]
        assert tokens == expected_tokens

    def test_tokenize_handles_special_cases(self):
        sentence = u"mr. and mrs. jones, etc., went to, e.g., the store"
        expected_tokens = [u"mr.", u"and", u"mrs.", u"jones", u",", u"etc.", u",", u"went", u"to", u",",
                           u"e.g.", u",", u"the", u"store"]
        tokens = [t.text for t in self.word_splitter.split_words(sentence)]
        assert tokens == expected_tokens


class TestLettersDigitsWordSplitter(AllenNlpTestCase):
    def setUp(self):
        super(TestLettersDigitsWordSplitter, self).setUp()
        self.word_splitter = LettersDigitsWordSplitter()

    def test_tokenize_handles_complex_punctuation(self):
        sentence = u"this (sentence) has 'crazy' \"punctuation\"."
        expected_tokens = [u"this", u"(", u"sentence", u")", u"has", u"'", u"crazy", u"'", u'"',
                           u"punctuation", u'"', u"."]
        tokens = [t.text for t in self.word_splitter.split_words(sentence)]
        assert tokens == expected_tokens

    def test_tokenize_handles_unicode_letters(self):
        sentence = u"HAL9000   and    Ångström"
        expected_tokens = [Token(u"HAL", 0), Token(u"9000", 3), Token(u"and", 10), Token(u"Ångström", 17)]
        tokens = self.word_splitter.split_words(sentence)
        assert [t.text for t in tokens] == [t.text for t in expected_tokens]
        assert [t.idx for t in tokens] == [t.idx for t in expected_tokens]

    def test_tokenize_handles_splits_all_punctuation(self):
        sentence = u"wouldn't.[have] -3.45(m^2)"
        expected_tokens = [u"wouldn", u"'", u"t", u".", u"[", u"have", u"]", u"-", u"3",
                           u".", u"45", u"(", u"m", u"^", u"2", u")"]
        tokens = [t.text for t in self.word_splitter.split_words(sentence)]
        assert tokens == expected_tokens


class TestSpacyWordSplitter(AllenNlpTestCase):
    def setUp(self):
        super(TestSpacyWordSplitter, self).setUp()
        self.word_splitter = SpacyWordSplitter()

    def test_tokenize_handles_complex_punctuation(self):
        sentence = u"this (sentence) has 'crazy' \"punctuation\"."
        expected_tokens = [u"this", u"(", u"sentence", u")", u"has", u"'", u"crazy", u"'", u'"',
                           u"punctuation", u'"', u"."]
        tokens = self.word_splitter.split_words(sentence)
        token_text = [t.text for t in tokens]
        assert token_text == expected_tokens
        for token in tokens:
            start = token.idx
            end = start + len(token.text)
            assert sentence[start:end] == token.text

    def test_tokenize_handles_contraction(self):
        # note that "would've" is kept together, while "ain't" is not.
        sentence = u"it ain't joe's problem; would been yesterday"
        expected_tokens = [u"it", u"ai", u"n't", u"joe", u"'s", u"problem", u";", u"would", u"been",
                           u"yesterday"]
        tokens = [t.text for t in self.word_splitter.split_words(sentence)]
        assert tokens == expected_tokens

    def test_tokenize_handles_multiple_contraction(self):
        sentence = u"wouldn't've"
        expected_tokens = [u"would", u"n't", u"'ve"]
        tokens = [t.text for t in self.word_splitter.split_words(sentence)]
        assert tokens == expected_tokens

    def test_tokenize_handles_final_apostrophe(self):
        sentence = u"the jones' house"
        expected_tokens = [u"the", u"jones", u"'", u"house"]
        tokens = [t.text for t in self.word_splitter.split_words(sentence)]
        assert tokens == expected_tokens

    def test_tokenize_removes_whitespace_tokens(self):
        sentence = u"the\n jones'   house  \x0b  55"
        expected_tokens = [u"the", u"jones", u"'", u"house", u"55"]
        tokens = [t.text for t in self.word_splitter.split_words(sentence)]
        assert tokens == expected_tokens

    def test_tokenize_handles_special_cases(self):
        # note that the etc. doesn't quite work --- we can special case this if we want.
        sentence = u"Mr. and Mrs. Jones, etc., went to, e.g., the store"
        expected_tokens = [u"Mr.", u"and", u"Mrs.", u"Jones", u",", u"etc", u".", u",", u"went", u"to", u",",
                           u"e.g.", u",", u"the", u"store"]
        tokens = [t.text for t in self.word_splitter.split_words(sentence)]
        assert tokens == expected_tokens

    def test_batch_tokenization(self):
        sentences = [u"This is     a sentence",
                     u"This isn't a sentence.",
                     u"This is the 3rd     sentence."
                     u"Here's the 'fourth' sentence."]
        batch_split = self.word_splitter.batch_split_words(sentences)
        separately_split = [self.word_splitter.split_words(sentence) for sentence in sentences]
        assert len(batch_split) == len(separately_split)
        for batch_sentence, separate_sentence in izip(batch_split, separately_split):
            assert len(batch_sentence) == len(separate_sentence)
            for batch_word, separate_word in izip(batch_sentence, separate_sentence):
                assert batch_word.text == separate_word.text



# pylint: disable=invalid-name,no-self-use,protected-access

from __future__ import absolute_import
import torch
from numpy.testing import assert_almost_equal
import numpy

from allennlp.common import Params
from allennlp.common.testing.test_case import AllenNlpTestCase
from allennlp.modules.attention.attention import Attention
from allennlp.modules.attention.dot_product_attention import DotProductAttention


class TestDotProductAttention(AllenNlpTestCase):

    def test_can_init_dot(self):
        legacy_attention = Attention.from_params(Params({u"type": u"dot_product"}))
        isinstance(legacy_attention, DotProductAttention)

    def test_dot_product_similarity(self):
        linear = DotProductAttention(normalize=False)
        output = linear(torch.FloatTensor([[0, 0, 0], [1, 1, 1]]),
                        torch.FloatTensor([[[1, 2, 3], [4, 5, 6]], [[7, 8, 9], [10, 11, 12]]]))

        assert_almost_equal(output.numpy(),
                            numpy.array([[0.0, 0.0], [24.0, 33.0]]), decimal=2)



# pylint: disable=no-self-use,invalid-name


from __future__ import division
from __future__ import absolute_import
import pytest

from allennlp.data.dataset_readers import SnliReader
from allennlp.common.util import ensure_list
from allennlp.common.testing import AllenNlpTestCase

class TestSnliReader(object):
    @pytest.mark.parametrize(u"lazy", (True, False))
    def test_read_from_file(self, lazy):
        reader = SnliReader(lazy=lazy)
        instances = reader.read(AllenNlpTestCase.FIXTURES_ROOT / u'data' / u'snli.jsonl')
        instances = ensure_list(instances)

        instance1 = {u"premise": [u"A", u"person", u"on", u"a", u"horse", u"jumps", u"over", u"a", u"broken",
                                 u"down", u"airplane", u"."],
                     u"hypothesis": [u"A", u"person", u"is", u"training", u"his", u"horse", u"for", u"a",
                                    u"competition", u"."],
                     u"label": u"neutral"}

        instance2 = {u"premise": [u"A", u"person", u"on", u"a", u"horse", u"jumps", u"over", u"a", u"broken",
                                 u"down", u"airplane", u"."],
                     u"hypothesis": [u"A", u"person", u"is", u"at", u"a", u"diner", u",", u"ordering", u"an",
                                    u"omelette", u"."],
                     u"label": u"contradiction"}
        instance3 = {u"premise": [u"A", u"person", u"on", u"a", u"horse", u"jumps", u"over", u"a", u"broken",
                                 u"down", u"airplane", u"."],
                     u"hypothesis": [u"A", u"person", u"is", u"outdoors", u",", u"on", u"a", u"horse", u"."],
                     u"label": u"entailment"}

        assert len(instances) == 3
        fields = instances[0].fields
        assert [t.text for t in fields[u"premise"].tokens] == instance1[u"premise"]
        assert [t.text for t in fields[u"hypothesis"].tokens] == instance1[u"hypothesis"]
        assert fields[u"label"].label == instance1[u"label"]
        fields = instances[1].fields
        assert [t.text for t in fields[u"premise"].tokens] == instance2[u"premise"]
        assert [t.text for t in fields[u"hypothesis"].tokens] == instance2[u"hypothesis"]
        assert fields[u"label"].label == instance2[u"label"]
        fields = instances[2].fields
        assert [t.text for t in fields[u"premise"].tokens] == instance3[u"premise"]
        assert [t.text for t in fields[u"hypothesis"].tokens] == instance3[u"hypothesis"]
        assert fields[u"label"].label == instance3[u"label"]





from __future__ import division
from __future__ import absolute_import
import json

from allennlp.common.testing import AllenNlpTestCase
from allennlp.semparse.worlds.atis_world import AtisWorld
from io import open

class TestAtisWorld(AllenNlpTestCase):
    def setUp(self):
        super(TestAtisWorld, self).setUp()
        test_filename = self.FIXTURES_ROOT / u"data" / u"atis" / u"sample.json"
        self.data = open(test_filename).readlines()

    def test_atis_global_actions(self): # pylint: disable=no-self-use
        world = AtisWorld([])
        valid_actions = world.valid_actions

        assert set(valid_actions.keys()) == set([u'agg',
                                             u'agg_func',
                                             u'agg_results',
                                             u'biexpr',
                                             u'binaryop',
                                             u'boolean',
                                             u'col_ref',
                                             u'col_refs',
                                             u'condition',
                                             u'conditions',
                                             u'conj',
                                             u'distinct',
                                             u'in_clause',
                                             u'number',
                                             u'pos_value',
                                             u'query',
                                             u'select_results',
                                             u'statement',
                                             u'string',
                                             u'table_name',
                                             u'table_refs',
                                             u'ternaryexpr',
                                             u'value',
                                             u'where_clause'])

        assert set(valid_actions[u'statement']) == set([u'statement -> [query, ";"]'])
        assert set(valid_actions[u'query']) ==\
                set([u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
                 u'where_clause, ")"]',
                 u'query -> ["SELECT", distinct, select_results, "FROM", table_refs, '
                 u'where_clause]'])
        assert set(valid_actions[u'select_results']) ==\
                set([u'select_results -> [agg]', u'select_results -> [col_refs]'])
        assert set(valid_actions[u'agg']) ==\
                set([u'agg -> [agg_func, "(", col_ref, ")"]'])
        assert set(valid_actions[u'agg_func']) ==\
                set([u'agg_func -> ["COUNT"]',
                 u'agg_func -> ["MAX"]',
                 u'agg_func -> ["MIN"]'])
        assert set(valid_actions[u'col_refs']) ==\
                set([u'col_refs -> [col_ref]', u'col_refs -> [col_ref, ",", col_refs]'])
        assert set(valid_actions[u'table_refs']) ==\
                set([u'table_refs -> [table_name]', u'table_refs -> [table_name, ",", table_refs]'])
        assert set(valid_actions[u'where_clause']) ==\
                set([u'where_clause -> ["WHERE", "(", conditions, ")"]',
                 u'where_clause -> ["WHERE", conditions]'])
        assert set(valid_actions[u'conditions']) ==\
                set([u'conditions -> ["(", conditions, ")", conj, conditions]',
                 u'conditions -> ["(", conditions, ")"]',
                 u'conditions -> ["NOT", conditions]',
                 u'conditions -> [condition, conj, "(", conditions, ")"]',
                 u'conditions -> [condition, conj, conditions]',
                 u'conditions -> [condition]'])
        assert set(valid_actions[u'condition']) ==\
                set([u'condition -> [biexpr]',
                 u'condition -> [in_clause]',
                 u'condition -> [ternaryexpr]'])
        assert set(valid_actions[u'in_clause']) ==\
                set([u'in_clause -> [col_ref, "IN", query]'])
        assert set(valid_actions[u'biexpr']) ==\
                set([u'biexpr -> [col_ref, "LIKE", string]',
                 u'biexpr -> [col_ref, binaryop, value]',
                 u'biexpr -> [value, binaryop, value]'])
        assert set(valid_actions[u'binaryop']) ==\
                set([u'binaryop -> ["*"]',
                 u'binaryop -> ["+"]',
                 u'binaryop -> ["-"]',
                 u'binaryop -> ["/"]',
                 u'binaryop -> ["<"]',
                 u'binaryop -> ["<="]',
                 u'binaryop -> ["="]',
                 u'binaryop -> [">"]',
                 u'binaryop -> [">="]',
                 u'binaryop -> ["IS"]'])
        assert set(valid_actions[u'ternaryexpr']) ==\
                set([u'ternaryexpr -> [col_ref, "BETWEEN", value, "AND", value]',
                 u'ternaryexpr -> [col_ref, "NOT", "BETWEEN", value, "AND", value]'])
        assert set(valid_actions[u'value']) ==\
                set([u'value -> ["NOT", pos_value]',
                 u'value -> [pos_value]'])
        assert set(valid_actions[u'pos_value']) ==\
                set([u'pos_value -> ["ALL", query]',
                 u'pos_value -> ["ANY", query]',
                 u'pos_value -> ["NULL"]',
                 u'pos_value -> [agg_results]',
                 u'pos_value -> [boolean]',
                 u'pos_value -> [col_ref]',
                 u'pos_value -> [number]',
                 u'pos_value -> [string]'])
        assert set(valid_actions[u'agg_results']) ==\
                set([(u'agg_results -> ["(", "SELECT", distinct, agg, "FROM", table_name, '
                  u'where_clause, ")"]'),
                 u'agg_results -> ["SELECT", distinct, agg, "FROM", table_name, where_clause]'])
        assert set(valid_actions[u'boolean']) ==\
                set([u'boolean -> ["true"]', u'boolean -> ["false"]'])
        assert set(valid_actions[u'conj']) ==\
                set([u'conj -> ["OR"]', u'conj -> ["AND"]'])
        assert set(valid_actions[u'distinct']) ==\
               set([u'distinct -> [""]', u'distinct -> ["DISTINCT"]'])
        assert set(valid_actions[u'number']) ==\
                set([u'number -> ["0"]',
                 u'number -> ["1"]'])
        assert set(valid_actions[u'string']) == set()
        assert set(valid_actions[u'col_ref']) ==\
                set([u'col_ref -> ["*"]',
                 u'col_ref -> ["aircraft", ".", "aircraft_code"]',
                 u'col_ref -> ["aircraft", ".", "aircraft_description"]',
                 u'col_ref -> ["aircraft", ".", "basic_type"]',
                 u'col_ref -> ["aircraft", ".", "manufacturer"]',
                 u'col_ref -> ["aircraft", ".", "pressurized"]',
                 u'col_ref -> ["aircraft", ".", "propulsion"]',
                 u'col_ref -> ["aircraft", ".", "wide_body"]',
                 u'col_ref -> ["airline", ".", "airline_code"]',
                 u'col_ref -> ["airline", ".", "airline_name"]',
                 u'col_ref -> ["airport", ".", "airport_code"]',
                 u'col_ref -> ["airport", ".", "airport_location"]',
                 u'col_ref -> ["airport", ".", "airport_name"]',
                 u'col_ref -> ["airport", ".", "country_name"]',
                 u'col_ref -> ["airport", ".", "minimum_connect_time"]',
                 u'col_ref -> ["airport", ".", "state_code"]',
                 u'col_ref -> ["airport", ".", "time_zone_code"]',
                 u'col_ref -> ["airport_service", ".", "airport_code"]',
                 u'col_ref -> ["airport_service", ".", "city_code"]',
                 u'col_ref -> ["airport_service", ".", "direction"]',
                 u'col_ref -> ["airport_service", ".", "miles_distant"]',
                 u'col_ref -> ["airport_service", ".", "minutes_distant"]',
                 u'col_ref -> ["city", ".", "city_code"]',
                 u'col_ref -> ["city", ".", "city_name"]',
                 u'col_ref -> ["city", ".", "country_name"]',
                 u'col_ref -> ["city", ".", "state_code"]',
                 u'col_ref -> ["city", ".", "time_zone_code"]',
                 u'col_ref -> ["class_of_service", ".", "booking_class"]',
                 u'col_ref -> ["class_of_service", ".", "class_description"]',
                 u'col_ref -> ["class_of_service", ".", "rank"]',
                 u'col_ref -> ["date_day", ".", "day_name"]',
                 u'col_ref -> ["date_day", ".", "day_number"]',
                 u'col_ref -> ["date_day", ".", "month_number"]',
                 u'col_ref -> ["date_day", ".", "year"]',
                 u'col_ref -> ["days", ".", "day_name"]',
                 u'col_ref -> ["days", ".", "days_code"]',
                 u'col_ref -> ["equipment_sequence", ".", "aircraft_code"]',
                 u'col_ref -> ["equipment_sequence", ".", "aircraft_code_sequence"]',
                 u'col_ref -> ["fare", ".", "fare_airline"]',
                 u'col_ref -> ["fare", ".", "fare_basis_code"]',
                 u'col_ref -> ["fare", ".", "fare_id"]',
                 u'col_ref -> ["fare", ".", "from_airport"]',
                 u'col_ref -> ["fare", ".", "one_direction_cost"]',
                 u'col_ref -> ["fare", ".", "restriction_code"]',
                 u'col_ref -> ["fare", ".", "round_trip_cost"]',
                 u'col_ref -> ["fare", ".", "round_trip_required"]',
                 u'col_ref -> ["fare", ".", "to_airport"]',
                 u'col_ref -> ["fare_basis", ".", "basis_days"]',
                 u'col_ref -> ["fare_basis", ".", "booking_class"]',
                 u'col_ref -> ["fare_basis", ".", "class_type"]',
                 u'col_ref -> ["fare_basis", ".", "discounted"]',
                 u'col_ref -> ["fare_basis", ".", "economy"]',
                 u'col_ref -> ["fare_basis", ".", "fare_basis_code"]',
                 u'col_ref -> ["fare_basis", ".", "night"]',
                 u'col_ref -> ["fare_basis", ".", "premium"]',
                 u'col_ref -> ["fare_basis", ".", "season"]',
                 u'col_ref -> ["flight", ".", "aircraft_code_sequence"]',
                 u'col_ref -> ["flight", ".", "airline_code"]',
                 u'col_ref -> ["flight", ".", "airline_flight"]',
                 u'col_ref -> ["flight", ".", "arrival_time"]',
                 u'col_ref -> ["flight", ".", "connections"]',
                 u'col_ref -> ["flight", ".", "departure_time"]',
                 u'col_ref -> ["flight", ".", "dual_carrier"]',
                 u'col_ref -> ["flight", ".", "flight_days"]',
                 u'col_ref -> ["flight", ".", "flight_id"]',
                 u'col_ref -> ["flight", ".", "flight_number"]',
                 u'col_ref -> ["flight", ".", "from_airport"]',
                 u'col_ref -> ["flight", ".", "meal_code"]',
                 u'col_ref -> ["flight", ".", "stops"]',
                 u'col_ref -> ["flight", ".", "time_elapsed"]',
                 u'col_ref -> ["flight", ".", "to_airport"]',
                 u'col_ref -> ["flight_fare", ".", "fare_id"]',
                 u'col_ref -> ["flight_fare", ".", "flight_id"]',
                 u'col_ref -> ["flight_leg", ".", "flight_id"]',
                 u'col_ref -> ["flight_leg", ".", "leg_flight"]',
                 u'col_ref -> ["flight_leg", ".", "leg_number"]',
                 u'col_ref -> ["flight_stop", ".", "arrival_airline"]',
                 u'col_ref -> ["flight_stop", ".", "arrival_flight_number"]',
                 u'col_ref -> ["flight_stop", ".", "arrival_time"]',
                 u'col_ref -> ["flight_stop", ".", "departure_airline"]',
                 u'col_ref -> ["flight_stop", ".", "departure_flight_number"]',
                 u'col_ref -> ["flight_stop", ".", "departure_time"]',
                 u'col_ref -> ["flight_stop", ".", "flight_id"]',
                 u'col_ref -> ["flight_stop", ".", "stop_airport"]',
                 u'col_ref -> ["flight_stop", ".", "stop_days"]',
                 u'col_ref -> ["flight_stop", ".", "stop_number"]',
                 u'col_ref -> ["flight_stop", ".", "stop_time"]',
                 u'col_ref -> ["food_service", ".", "compartment"]',
                 u'col_ref -> ["food_service", ".", "meal_code"]',
                 u'col_ref -> ["food_service", ".", "meal_description"]',
                 u'col_ref -> ["food_service", ".", "meal_number"]',
                 u'col_ref -> ["ground_service", ".", "airport_code"]',
                 u'col_ref -> ["ground_service", ".", "city_code"]',
                 u'col_ref -> ["ground_service", ".", "ground_fare"]',
                 u'col_ref -> ["ground_service", ".", "transport_type"]',
                 u'col_ref -> ["month", ".", "month_name"]',
                 u'col_ref -> ["month", ".", "month_number"]',
                 u'col_ref -> ["restriction", ".", "advance_purchase"]',
                 u'col_ref -> ["restriction", ".", "application"]',
                 u'col_ref -> ["restriction", ".", "maximum_stay"]',
                 u'col_ref -> ["restriction", ".", "minimum_stay"]',
                 u'col_ref -> ["restriction", ".", "no_discounts"]',
                 u'col_ref -> ["restriction", ".", "restriction_code"]',
                 u'col_ref -> ["restriction", ".", "saturday_stay_required"]',
                 u'col_ref -> ["restriction", ".", "stopovers"]',
                 u'col_ref -> ["state", ".", "country_name"]',
                 u'col_ref -> ["state", ".", "state_code"]',
                 u'col_ref -> ["state", ".", "state_name"]'])

        assert set(valid_actions[u'table_name']) ==\
                set([u'table_name -> ["aircraft"]',
                 u'table_name -> ["airline"]',
                 u'table_name -> ["airport"]',
                 u'table_name -> ["airport_service"]',
                 u'table_name -> ["city"]',
                 u'table_name -> ["class_of_service"]',
                 u'table_name -> ["date_day"]',
                 u'table_name -> ["days"]',
                 u'table_name -> ["equipment_sequence"]',
                 u'table_name -> ["fare"]',
                 u'table_name -> ["fare_basis"]',
                 u'table_name -> ["flight"]',
                 u'table_name -> ["flight_fare"]',
                 u'table_name -> ["flight_leg"]',
                 u'table_name -> ["flight_stop"]',
                 u'table_name -> ["food_service"]',
                 u'table_name -> ["ground_service"]',
                 u'table_name -> ["month"]',
                 u'table_name -> ["restriction"]',
                 u'table_name -> ["state"]'])

    def test_atis_local_actions(self): # pylint: disable=no-self-use
        # Check if the triggers activate correcty
        world = AtisWorld([u"show me the flights from denver at 12 o'clock"])
        assert set(world.valid_actions[u'number']) ==\
            set([u'number -> ["0"]',
             u'number -> ["1"]',
             u'number -> ["1200"]',
             u'number -> ["2400"]'])

        assert set(world.valid_actions[u'string']) ==\
                set([u'string -> ["\'DENVER\'"]',
                 u'string -> ["\'DDEN\'"]',
                 u'string -> ["\'AT\'"]'])

        world = AtisWorld([u"show me the flights from denver at 12 o'clock",
                           u"show me the delta or united flights in afternoon"])

        assert set(world.valid_actions[u'number']) ==\
                set([u'number -> ["0"]',
                 u'number -> ["1"]',
                 u'number -> ["1800"]',
                 u'number -> ["1200"]',
                 u'number -> ["2400"]'])

        assert set(world.valid_actions[u'string']) ==\
                set([u'string -> ["\'DENVER\'"]',
                 u'string -> ["\'DDEN\'"]',
                 u'string -> ["\'AT\'"]',
                 u'string -> ["\'DL\'"]',
                 u'string -> ["\'UA\'"]',
                 u'string -> ["\'IN\'"]'])

        world = AtisWorld([u"i would like one coach reservation for \
                          may ninth from pittsburgh to atlanta leaving \
                          pittsburgh before 10 o'clock in morning 1991 \
                          august twenty sixth"])

        assert set(world.valid_actions[u'number']) ==\
                set([u'number -> ["0"]',
                 u'number -> ["1"]',
                 u'number -> ["9"]',
                 u'number -> ["8"]',
                 u'number -> ["6"]',
                 u'number -> ["5"]',
                 u'number -> ["26"]',
                 u'number -> ["2200"]',
                 u'number -> ["1991"]',
                 u'number -> ["1200"]',
                 u'number -> ["1000"]'])

        assert set(world.valid_actions[u'string']) ==\
                set([u'string -> ["\'COACH\'"]',
                 u'string -> ["\'PITTSBURGH\'"]',
                 u'string -> ["\'PIT\'"]',
                 u'string -> ["\'PPIT\'"]',
                 u'string -> ["\'ATLANTA\'"]',
                 u'string -> ["\'ATL\'"]',
                 u'string -> ["\'MATL\'"]',
                 u'string -> ["\'IN\'"]',
                 u'string -> ["\'MONDAY\'"]'])


    def test_atis_simple_action_sequence(self): # pylint: disable=no-self-use
        world = AtisWorld([(u"give me all flights from boston to "
                            u"philadelphia next week arriving after lunch")])
        action_sequence = world.get_action_sequence((u"(SELECT DISTINCT city . city_code , city . city_name "
                                                     u"FROM city WHERE ( city.city_name = 'BOSTON' ) );"))
        assert action_sequence == [u'statement -> [query, ";"]',
                                   u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
                                   u'where_clause, ")"]',
                                   u'where_clause -> ["WHERE", "(", conditions, ")"]',
                                   u'conditions -> [condition]',
                                   u'condition -> [biexpr]',
                                   u'biexpr -> [col_ref, binaryop, value]',
                                   u'value -> [pos_value]',
                                   u'pos_value -> [string]',
                                   u'string -> ["\'BOSTON\'"]',
                                   u'binaryop -> ["="]',
                                   u'col_ref -> ["city", ".", "city_name"]',
                                   u'table_refs -> [table_name]',
                                   u'table_name -> ["city"]',
                                   u'select_results -> [col_refs]',
                                   u'col_refs -> [col_ref, ",", col_refs]',
                                   u'col_refs -> [col_ref]',
                                   u'col_ref -> ["city", ".", "city_name"]',
                                   u'col_ref -> ["city", ".", "city_code"]',
                                   u'distinct -> ["DISTINCT"]']

        action_sequence = world.get_action_sequence((u"( SELECT airport_service . airport_code "
                                                     u"FROM airport_service "
                                                     u"WHERE airport_service . city_code IN ( "
                                                     u"SELECT city . city_code FROM city "
                                                     u"WHERE city.city_name = 'BOSTON' ) ) ;"))

        assert action_sequence == [u'statement -> [query, ";"]',
                                   u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
                                   u'where_clause, ")"]',
                                   u'where_clause -> ["WHERE", conditions]',
                                   u'conditions -> [condition]',
                                   u'condition -> [in_clause]',
                                   u'in_clause -> [col_ref, "IN", query]',
                                   u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
                                   u'where_clause, ")"]',
                                   u'where_clause -> ["WHERE", conditions]',
                                   u'conditions -> [condition]',
                                   u'condition -> [biexpr]',
                                   u'biexpr -> [col_ref, binaryop, value]',
                                   u'value -> [pos_value]',
                                   u'pos_value -> [string]',
                                   u'string -> ["\'BOSTON\'"]',
                                   u'binaryop -> ["="]',
                                   u'col_ref -> ["city", ".", "city_name"]',
                                   u'table_refs -> [table_name]',
                                   u'table_name -> ["city"]',
                                   u'select_results -> [col_refs]',
                                   u'col_refs -> [col_ref]',
                                   u'col_ref -> ["city", ".", "city_code"]',
                                   u'distinct -> [""]',
                                   u'col_ref -> ["airport_service", ".", "city_code"]',
                                   u'table_refs -> [table_name]',
                                   u'table_name -> ["airport_service"]',
                                   u'select_results -> [col_refs]',
                                   u'col_refs -> [col_ref]',
                                   u'col_ref -> ["airport_service", ".", "airport_code"]',
                                   u'distinct -> [""]']

        action_sequence = world.get_action_sequence((u"( SELECT airport_service . airport_code "
                                                     u"FROM airport_service WHERE airport_service . city_code IN "
                                                     u"( SELECT city . city_code FROM city "
                                                     u"WHERE city.city_name = 'BOSTON' ) AND 1 = 1) ;"))

        assert action_sequence ==\
                [u'statement -> [query, ";"]',
                 u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
                 u'where_clause, ")"]',
                 u'where_clause -> ["WHERE", conditions]',
                 u'conditions -> [condition, conj, conditions]',
                 u'conditions -> [condition]',
                 u'condition -> [biexpr]',
                 u'biexpr -> [value, binaryop, value]',
                 u'value -> [pos_value]',
                 u'pos_value -> [number]',
                 u'number -> ["1"]',
                 u'binaryop -> ["="]',
                 u'value -> [pos_value]',
                 u'pos_value -> [number]',
                 u'number -> ["1"]',
                 u'conj -> ["AND"]',
                 u'condition -> [in_clause]',
                 u'in_clause -> [col_ref, "IN", query]',
                 u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
                 u'where_clause, ")"]',
                 u'where_clause -> ["WHERE", conditions]',
                 u'conditions -> [condition]',
                 u'condition -> [biexpr]',
                 u'biexpr -> [col_ref, binaryop, value]',
                 u'value -> [pos_value]',
                 u'pos_value -> [string]',
                 u'string -> ["\'BOSTON\'"]',
                 u'binaryop -> ["="]',
                 u'col_ref -> ["city", ".", "city_name"]',
                 u'table_refs -> [table_name]',
                 u'table_name -> ["city"]',
                 u'select_results -> [col_refs]',
                 u'col_refs -> [col_ref]',
                 u'col_ref -> ["city", ".", "city_code"]',
                 u'distinct -> [""]',
                 u'col_ref -> ["airport_service", ".", "city_code"]',
                 u'table_refs -> [table_name]',
                 u'table_name -> ["airport_service"]',
                 u'select_results -> [col_refs]',
                 u'col_refs -> [col_ref]',
                 u'col_ref -> ["airport_service", ".", "airport_code"]',
                 u'distinct -> [""]']

        world = AtisWorld([(u"give me all flights from boston to "
                            u"philadelphia next week arriving after lunch")])
        action_sequence = world.get_action_sequence((u"( SELECT DISTINCT flight.flight_id "
                                                     u"FROM flight WHERE "
                                                     u"( flight . from_airport IN "
                                                     u"( SELECT airport_service . airport_code "
                                                     u"FROM airport_service WHERE airport_service . city_code IN "
                                                     u"( SELECT city . city_code "
                                                     u"FROM city "
                                                     u"WHERE city.city_name = 'BOSTON' )))) ;"))

        assert action_sequence ==\
            [u'statement -> [query, ";"]',
             u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
             u'where_clause, ")"]',
             u'where_clause -> ["WHERE", "(", conditions, ")"]',
             u'conditions -> [condition]',
             u'condition -> [in_clause]',
             u'in_clause -> [col_ref, "IN", query]',
             u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
             u'where_clause, ")"]',
             u'where_clause -> ["WHERE", conditions]',
             u'conditions -> [condition]',
             u'condition -> [in_clause]',
             u'in_clause -> [col_ref, "IN", query]',
             u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
             u'where_clause, ")"]',
             u'where_clause -> ["WHERE", conditions]',
             u'conditions -> [condition]',
             u'condition -> [biexpr]',
             u'biexpr -> [col_ref, binaryop, value]',
             u'value -> [pos_value]',
             u'pos_value -> [string]',
             u'string -> ["\'BOSTON\'"]',
             u'binaryop -> ["="]',
             u'col_ref -> ["city", ".", "city_name"]',
             u'table_refs -> [table_name]',
             u'table_name -> ["city"]',
             u'select_results -> [col_refs]',
             u'col_refs -> [col_ref]',
             u'col_ref -> ["city", ".", "city_code"]',
             u'distinct -> [""]',
             u'col_ref -> ["airport_service", ".", "city_code"]',
             u'table_refs -> [table_name]',
             u'table_name -> ["airport_service"]',
             u'select_results -> [col_refs]',
             u'col_refs -> [col_ref]',
             u'col_ref -> ["airport_service", ".", "airport_code"]',
             u'distinct -> [""]',
             u'col_ref -> ["flight", ".", "from_airport"]',
             u'table_refs -> [table_name]',
             u'table_name -> ["flight"]',
             u'select_results -> [col_refs]',
             u'col_refs -> [col_ref]',
             u'col_ref -> ["flight", ".", "flight_id"]',
             u'distinct -> ["DISTINCT"]']

    def test_atis_long_action_sequence(self): # pylint: disable=no-self-use
        world = AtisWorld([(u"what is the earliest flight in morning "
                            u"1993 june fourth from boston to pittsburgh")])
        action_sequence = world.get_action_sequence(u"( SELECT DISTINCT flight.flight_id "
                                                    u"FROM flight "
                                                    u"WHERE ( flight.departure_time = ( "
                                                    u"SELECT MIN ( flight.departure_time ) "
                                                    u"FROM flight "
                                                    u"WHERE ( flight.departure_time BETWEEN 0 AND 1200 AND "
                                                    u"( flight . from_airport IN ( "
                                                    u"SELECT airport_service . airport_code "
                                                    u"FROM airport_service WHERE airport_service . city_code "
                                                    u"IN ( "
                                                    u"SELECT city . city_code "
                                                    u"FROM city WHERE city.city_name = 'BOSTON' )) "
                                                    u"AND flight . to_airport IN ( "
                                                    u"SELECT airport_service . airport_code "
                                                    u"FROM airport_service "
                                                    u"WHERE airport_service . city_code IN ( "
                                                    u"SELECT city . city_code "
                                                    u"FROM city "
                                                    u"WHERE city.city_name = 'PITTSBURGH' )) ) ) ) AND "
                                                    u"( flight.departure_time BETWEEN 0 AND 1200 AND "
                                                    u"( flight . from_airport IN ( "
                                                    u"SELECT airport_service . airport_code "
                                                    u"FROM airport_service "
                                                    u"WHERE airport_service . city_code IN ( "
                                                    u"SELECT city . city_code "
                                                    u"FROM city WHERE city.city_name = 'BOSTON' )) "
                                                    u"AND flight . to_airport IN ( "
                                                    u"SELECT airport_service . airport_code "
                                                    u"FROM airport_service WHERE airport_service . city_code IN ( "
                                                    u"SELECT city . city_code "
                                                    u"FROM city "
                                                    u"WHERE city.city_name = 'PITTSBURGH' )) ) ) )   ) ;")
        assert action_sequence ==\
            [u'statement -> [query, ";"]',
             u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
             u'where_clause, ")"]',
             u'where_clause -> ["WHERE", "(", conditions, ")"]',
             u'conditions -> [condition, conj, conditions]',
             u'conditions -> ["(", conditions, ")"]',
             u'conditions -> [condition, conj, conditions]',
             u'conditions -> ["(", conditions, ")"]',
             u'conditions -> [condition, conj, conditions]',
             u'conditions -> [condition]',
             u'condition -> [in_clause]',
             u'in_clause -> [col_ref, "IN", query]',
             u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
             u'where_clause, ")"]',
             u'where_clause -> ["WHERE", conditions]',
             u'conditions -> [condition]',
             u'condition -> [in_clause]',
             u'in_clause -> [col_ref, "IN", query]',
             u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
             u'where_clause, ")"]',
             u'where_clause -> ["WHERE", conditions]',
             u'conditions -> [condition]',
             u'condition -> [biexpr]',
             u'biexpr -> [col_ref, binaryop, value]',
             u'value -> [pos_value]',
             u'pos_value -> [string]',
             u'string -> ["\'PITTSBURGH\'"]',
             u'binaryop -> ["="]',
             u'col_ref -> ["city", ".", "city_name"]',
             u'table_refs -> [table_name]',
             u'table_name -> ["city"]',
             u'select_results -> [col_refs]',
             u'col_refs -> [col_ref]',
             u'col_ref -> ["city", ".", "city_code"]',
             u'distinct -> [""]',
             u'col_ref -> ["airport_service", ".", "city_code"]',
             u'table_refs -> [table_name]',
             u'table_name -> ["airport_service"]',
             u'select_results -> [col_refs]',
             u'col_refs -> [col_ref]',
             u'col_ref -> ["airport_service", ".", "airport_code"]',
             u'distinct -> [""]',
             u'col_ref -> ["flight", ".", "to_airport"]',
             u'conj -> ["AND"]',
             u'condition -> [in_clause]',
             u'in_clause -> [col_ref, "IN", query]',
             u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
             u'where_clause, ")"]',
             u'where_clause -> ["WHERE", conditions]',
             u'conditions -> [condition]',
             u'condition -> [in_clause]',
             u'in_clause -> [col_ref, "IN", query]',
             u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
             u'where_clause, ")"]',
             u'where_clause -> ["WHERE", conditions]',
             u'conditions -> [condition]',
             u'condition -> [biexpr]',
             u'biexpr -> [col_ref, binaryop, value]',
             u'value -> [pos_value]',
             u'pos_value -> [string]',
             u'string -> ["\'BOSTON\'"]',
             u'binaryop -> ["="]',
             u'col_ref -> ["city", ".", "city_name"]',
             u'table_refs -> [table_name]',
             u'table_name -> ["city"]',
             u'select_results -> [col_refs]',
             u'col_refs -> [col_ref]',
             u'col_ref -> ["city", ".", "city_code"]',
             u'distinct -> [""]',
             u'col_ref -> ["airport_service", ".", "city_code"]',
             u'table_refs -> [table_name]',
             u'table_name -> ["airport_service"]',
             u'select_results -> [col_refs]',
             u'col_refs -> [col_ref]',
             u'col_ref -> ["airport_service", ".", "airport_code"]',
             u'distinct -> [""]',
             u'col_ref -> ["flight", ".", "from_airport"]',
             u'conj -> ["AND"]',
             u'condition -> [ternaryexpr]',
             u'ternaryexpr -> [col_ref, "BETWEEN", value, "AND", value]',
             u'value -> [pos_value]',
             u'pos_value -> [number]',
             u'number -> ["1200"]',
             u'value -> [pos_value]',
             u'pos_value -> [number]',
             u'number -> ["0"]',
             u'col_ref -> ["flight", ".", "departure_time"]',
             u'conj -> ["AND"]',
             u'condition -> [biexpr]',
             u'biexpr -> [col_ref, binaryop, value]',
             u'value -> [pos_value]',
             u'pos_value -> [agg_results]',
             u'agg_results -> ["(", "SELECT", distinct, agg, "FROM", table_name, '
             u'where_clause, ")"]',
             u'where_clause -> ["WHERE", "(", conditions, ")"]',
             u'conditions -> [condition, conj, conditions]',
             u'conditions -> ["(", conditions, ")"]',
             u'conditions -> [condition, conj, conditions]',
             u'conditions -> [condition]',
             u'condition -> [in_clause]',
             u'in_clause -> [col_ref, "IN", query]',
             u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
             u'where_clause, ")"]',
             u'where_clause -> ["WHERE", conditions]',
             u'conditions -> [condition]',
             u'condition -> [in_clause]',
             u'in_clause -> [col_ref, "IN", query]',
             u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
             u'where_clause, ")"]',
             u'where_clause -> ["WHERE", conditions]',
             u'conditions -> [condition]',
             u'condition -> [biexpr]',
             u'biexpr -> [col_ref, binaryop, value]',
             u'value -> [pos_value]',
             u'pos_value -> [string]',
             u'string -> ["\'PITTSBURGH\'"]',
             u'binaryop -> ["="]',
             u'col_ref -> ["city", ".", "city_name"]',
             u'table_refs -> [table_name]',
             u'table_name -> ["city"]',
             u'select_results -> [col_refs]',
             u'col_refs -> [col_ref]',
             u'col_ref -> ["city", ".", "city_code"]',
             u'distinct -> [""]',
             u'col_ref -> ["airport_service", ".", "city_code"]',
             u'table_refs -> [table_name]',
             u'table_name -> ["airport_service"]',
             u'select_results -> [col_refs]',
             u'col_refs -> [col_ref]',
             u'col_ref -> ["airport_service", ".", "airport_code"]',
             u'distinct -> [""]',
             u'col_ref -> ["flight", ".", "to_airport"]',
             u'conj -> ["AND"]',
             u'condition -> [in_clause]',
             u'in_clause -> [col_ref, "IN", query]',
             u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
             u'where_clause, ")"]',
             u'where_clause -> ["WHERE", conditions]',
             u'conditions -> [condition]',
             u'condition -> [in_clause]',
             u'in_clause -> [col_ref, "IN", query]',
             u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
             u'where_clause, ")"]',
             u'where_clause -> ["WHERE", conditions]',
             u'conditions -> [condition]',
             u'condition -> [biexpr]',
             u'biexpr -> [col_ref, binaryop, value]',
             u'value -> [pos_value]',
             u'pos_value -> [string]',
             u'string -> ["\'BOSTON\'"]',
             u'binaryop -> ["="]',
             u'col_ref -> ["city", ".", "city_name"]',
             u'table_refs -> [table_name]',
             u'table_name -> ["city"]',
             u'select_results -> [col_refs]',
             u'col_refs -> [col_ref]',
             u'col_ref -> ["city", ".", "city_code"]',
             u'distinct -> [""]',
             u'col_ref -> ["airport_service", ".", "city_code"]',
             u'table_refs -> [table_name]',
             u'table_name -> ["airport_service"]',
             u'select_results -> [col_refs]',
             u'col_refs -> [col_ref]',
             u'col_ref -> ["airport_service", ".", "airport_code"]',
             u'distinct -> [""]',
             u'col_ref -> ["flight", ".", "from_airport"]',
             u'conj -> ["AND"]',
             u'condition -> [ternaryexpr]',
             u'ternaryexpr -> [col_ref, "BETWEEN", value, "AND", value]',
             u'value -> [pos_value]',
             u'pos_value -> [number]',
             u'number -> ["1200"]',
             u'value -> [pos_value]',
             u'pos_value -> [number]',
             u'number -> ["0"]',
             u'col_ref -> ["flight", ".", "departure_time"]',
             u'table_name -> ["flight"]',
             u'agg -> [agg_func, "(", col_ref, ")"]',
             u'col_ref -> ["flight", ".", "departure_time"]',
             u'agg_func -> ["MIN"]',
             u'distinct -> [""]',
             u'binaryop -> ["="]',
             u'col_ref -> ["flight", ".", "departure_time"]',
             u'table_refs -> [table_name]',
             u'table_name -> ["flight"]',
             u'select_results -> [col_refs]',
             u'col_refs -> [col_ref]',
             u'col_ref -> ["flight", ".", "flight_id"]',
             u'distinct -> ["DISTINCT"]']


    def test_atis_from_json(self):
        line = json.loads(self.data[0])
        for utterance_idx in range(len(line[u'interaction'])):
            world = AtisWorld([interaction[u'utterance'] for
                               interaction in line[u'interaction'][:utterance_idx+1]])
            action_sequence = world.get_action_sequence(line[u'interaction'][utterance_idx][u'sql'])
            assert action_sequence is not None

    def test_all_possible_actions(self): # pylint: disable=no-self-use
        world = AtisWorld([(u"give me all flights from boston to "
                            u"philadelphia next week arriving after lunch")])
        possible_actions = world.all_possible_actions()

        assert possible_actions ==\
            [u'agg -> [agg_func, "(", col_ref, ")"]',
             u'agg_func -> ["COUNT"]',
             u'agg_func -> ["MAX"]',
             u'agg_func -> ["MIN"]',
             u'agg_results -> ["(", "SELECT", distinct, agg, "FROM", table_name, '
             u'where_clause, ")"]',
             u'agg_results -> ["SELECT", distinct, agg, "FROM", table_name, where_clause]',
             u'biexpr -> [col_ref, "LIKE", string]',
             u'biexpr -> [col_ref, binaryop, value]',
             u'biexpr -> [value, binaryop, value]',
             u'binaryop -> ["*"]',
             u'binaryop -> ["+"]',
             u'binaryop -> ["-"]',
             u'binaryop -> ["/"]',
             u'binaryop -> ["<"]',
             u'binaryop -> ["<="]',
             u'binaryop -> ["="]',
             u'binaryop -> [">"]',
             u'binaryop -> [">="]',
             u'binaryop -> ["IS"]',
             u'boolean -> ["false"]',
             u'boolean -> ["true"]',
             u'col_ref -> ["*"]',
             u'col_ref -> ["aircraft", ".", "aircraft_code"]',
             u'col_ref -> ["aircraft", ".", "aircraft_description"]',
             u'col_ref -> ["aircraft", ".", "basic_type"]',
             u'col_ref -> ["aircraft", ".", "manufacturer"]',
             u'col_ref -> ["aircraft", ".", "pressurized"]',
             u'col_ref -> ["aircraft", ".", "propulsion"]',
             u'col_ref -> ["aircraft", ".", "wide_body"]',
             u'col_ref -> ["airline", ".", "airline_code"]',
             u'col_ref -> ["airline", ".", "airline_name"]',
             u'col_ref -> ["airport", ".", "airport_code"]',
             u'col_ref -> ["airport", ".", "airport_location"]',
             u'col_ref -> ["airport", ".", "airport_name"]',
             u'col_ref -> ["airport", ".", "country_name"]',
             u'col_ref -> ["airport", ".", "minimum_connect_time"]',
             u'col_ref -> ["airport", ".", "state_code"]',
             u'col_ref -> ["airport", ".", "time_zone_code"]',
             u'col_ref -> ["airport_service", ".", "airport_code"]',
             u'col_ref -> ["airport_service", ".", "city_code"]',
             u'col_ref -> ["airport_service", ".", "direction"]',
             u'col_ref -> ["airport_service", ".", "miles_distant"]',
             u'col_ref -> ["airport_service", ".", "minutes_distant"]',
             u'col_ref -> ["city", ".", "city_code"]',
             u'col_ref -> ["city", ".", "city_name"]',
             u'col_ref -> ["city", ".", "country_name"]',
             u'col_ref -> ["city", ".", "state_code"]',
             u'col_ref -> ["city", ".", "time_zone_code"]',
             u'col_ref -> ["class_of_service", ".", "booking_class"]',
             u'col_ref -> ["class_of_service", ".", "class_description"]',
             u'col_ref -> ["class_of_service", ".", "rank"]',
             u'col_ref -> ["date_day", ".", "day_name"]',
             u'col_ref -> ["date_day", ".", "day_number"]',
             u'col_ref -> ["date_day", ".", "month_number"]',
             u'col_ref -> ["date_day", ".", "year"]',
             u'col_ref -> ["days", ".", "day_name"]',
             u'col_ref -> ["days", ".", "days_code"]',
             u'col_ref -> ["equipment_sequence", ".", "aircraft_code"]',
             u'col_ref -> ["equipment_sequence", ".", "aircraft_code_sequence"]',
             u'col_ref -> ["fare", ".", "fare_airline"]',
             u'col_ref -> ["fare", ".", "fare_basis_code"]',
             u'col_ref -> ["fare", ".", "fare_id"]',
             u'col_ref -> ["fare", ".", "from_airport"]',
             u'col_ref -> ["fare", ".", "one_direction_cost"]',
             u'col_ref -> ["fare", ".", "restriction_code"]',
             u'col_ref -> ["fare", ".", "round_trip_cost"]',
             u'col_ref -> ["fare", ".", "round_trip_required"]',
             u'col_ref -> ["fare", ".", "to_airport"]',
             u'col_ref -> ["fare_basis", ".", "basis_days"]',
             u'col_ref -> ["fare_basis", ".", "booking_class"]',
             u'col_ref -> ["fare_basis", ".", "class_type"]',
             u'col_ref -> ["fare_basis", ".", "discounted"]',
             u'col_ref -> ["fare_basis", ".", "economy"]',
             u'col_ref -> ["fare_basis", ".", "fare_basis_code"]',
             u'col_ref -> ["fare_basis", ".", "night"]',
             u'col_ref -> ["fare_basis", ".", "premium"]',
             u'col_ref -> ["fare_basis", ".", "season"]',
             u'col_ref -> ["flight", ".", "aircraft_code_sequence"]',
             u'col_ref -> ["flight", ".", "airline_code"]',
             u'col_ref -> ["flight", ".", "airline_flight"]',
             u'col_ref -> ["flight", ".", "arrival_time"]',
             u'col_ref -> ["flight", ".", "connections"]',
             u'col_ref -> ["flight", ".", "departure_time"]',
             u'col_ref -> ["flight", ".", "dual_carrier"]',
             u'col_ref -> ["flight", ".", "flight_days"]',
             u'col_ref -> ["flight", ".", "flight_id"]',
             u'col_ref -> ["flight", ".", "flight_number"]',
             u'col_ref -> ["flight", ".", "from_airport"]',
             u'col_ref -> ["flight", ".", "meal_code"]',
             u'col_ref -> ["flight", ".", "stops"]',
             u'col_ref -> ["flight", ".", "time_elapsed"]',
             u'col_ref -> ["flight", ".", "to_airport"]',
             u'col_ref -> ["flight_fare", ".", "fare_id"]',
             u'col_ref -> ["flight_fare", ".", "flight_id"]',
             u'col_ref -> ["flight_leg", ".", "flight_id"]',
             u'col_ref -> ["flight_leg", ".", "leg_flight"]',
             u'col_ref -> ["flight_leg", ".", "leg_number"]',
             u'col_ref -> ["flight_stop", ".", "arrival_airline"]',
             u'col_ref -> ["flight_stop", ".", "arrival_flight_number"]',
             u'col_ref -> ["flight_stop", ".", "arrival_time"]',
             u'col_ref -> ["flight_stop", ".", "departure_airline"]',
             u'col_ref -> ["flight_stop", ".", "departure_flight_number"]',
             u'col_ref -> ["flight_stop", ".", "departure_time"]',
             u'col_ref -> ["flight_stop", ".", "flight_id"]',
             u'col_ref -> ["flight_stop", ".", "stop_airport"]',
             u'col_ref -> ["flight_stop", ".", "stop_days"]',
             u'col_ref -> ["flight_stop", ".", "stop_number"]',
             u'col_ref -> ["flight_stop", ".", "stop_time"]',
             u'col_ref -> ["food_service", ".", "compartment"]',
             u'col_ref -> ["food_service", ".", "meal_code"]',
             u'col_ref -> ["food_service", ".", "meal_description"]',
             u'col_ref -> ["food_service", ".", "meal_number"]',
             u'col_ref -> ["ground_service", ".", "airport_code"]',
             u'col_ref -> ["ground_service", ".", "city_code"]',
             u'col_ref -> ["ground_service", ".", "ground_fare"]',
             u'col_ref -> ["ground_service", ".", "transport_type"]',
             u'col_ref -> ["month", ".", "month_name"]',
             u'col_ref -> ["month", ".", "month_number"]',
             u'col_ref -> ["restriction", ".", "advance_purchase"]',
             u'col_ref -> ["restriction", ".", "application"]',
             u'col_ref -> ["restriction", ".", "maximum_stay"]',
             u'col_ref -> ["restriction", ".", "minimum_stay"]',
             u'col_ref -> ["restriction", ".", "no_discounts"]',
             u'col_ref -> ["restriction", ".", "restriction_code"]',
             u'col_ref -> ["restriction", ".", "saturday_stay_required"]',
             u'col_ref -> ["restriction", ".", "stopovers"]',
             u'col_ref -> ["state", ".", "country_name"]',
             u'col_ref -> ["state", ".", "state_code"]',
             u'col_ref -> ["state", ".", "state_name"]',
             u'col_refs -> [col_ref, ",", col_refs]',
             u'col_refs -> [col_ref]',
             u'condition -> [biexpr]',
             u'condition -> [in_clause]',
             u'condition -> [ternaryexpr]',
             u'conditions -> ["(", conditions, ")", conj, conditions]',
             u'conditions -> ["(", conditions, ")"]',
             u'conditions -> ["NOT", conditions]',
             u'conditions -> [condition, conj, "(", conditions, ")"]',
             u'conditions -> [condition, conj, conditions]',
             u'conditions -> [condition]',
             u'conj -> ["AND"]',
             u'conj -> ["OR"]',
             u'distinct -> [""]',
             u'distinct -> ["DISTINCT"]',
             u'in_clause -> [col_ref, "IN", query]',
             u'number -> ["0"]',
             u'number -> ["1"]',
             u'number -> ["1200"]',
             u'number -> ["1400"]',
             u'number -> ["1800"]',
             u'pos_value -> ["ALL", query]',
             u'pos_value -> ["ANY", query]',
             u'pos_value -> ["NULL"]',
             u'pos_value -> [agg_results]',
             u'pos_value -> [boolean]',
             u'pos_value -> [col_ref]',
             u'pos_value -> [number]',
             u'pos_value -> [string]',
             u'query -> ["(", "SELECT", distinct, select_results, "FROM", table_refs, '
             u'where_clause, ")"]',
             u'query -> ["SELECT", distinct, select_results, "FROM", table_refs, '
             u'where_clause]',
             u'select_results -> [agg]',
             u'select_results -> [col_refs]',
             u'statement -> [query, ";"]',
             u'string -> ["\'BBOS\'"]',
             u'string -> ["\'BOS\'"]',
             u'string -> ["\'BOSTON\'"]',
             u'string -> ["\'LUNCH\'"]',
             u'string -> ["\'PHILADELPHIA\'"]',
             u'string -> ["\'PHL\'"]',
             u'string -> ["\'PPHL\'"]',
             u'table_name -> ["aircraft"]',
             u'table_name -> ["airline"]',
             u'table_name -> ["airport"]',
             u'table_name -> ["airport_service"]',
             u'table_name -> ["city"]',
             u'table_name -> ["class_of_service"]',
             u'table_name -> ["date_day"]',
             u'table_name -> ["days"]',
             u'table_name -> ["equipment_sequence"]',
             u'table_name -> ["fare"]',
             u'table_name -> ["fare_basis"]',
             u'table_name -> ["flight"]',
             u'table_name -> ["flight_fare"]',
             u'table_name -> ["flight_leg"]',
             u'table_name -> ["flight_stop"]',
             u'table_name -> ["food_service"]',
             u'table_name -> ["ground_service"]',
             u'table_name -> ["month"]',
             u'table_name -> ["restriction"]',
             u'table_name -> ["state"]',
             u'table_refs -> [table_name, ",", table_refs]',
             u'table_refs -> [table_name]',
             u'ternaryexpr -> [col_ref, "BETWEEN", value, "AND", value]',
             u'ternaryexpr -> [col_ref, "NOT", "BETWEEN", value, "AND", value]',
             u'value -> ["NOT", pos_value]',
             u'value -> [pos_value]',
             u'where_clause -> ["WHERE", "(", conditions, ")"]',
             u'where_clause -> ["WHERE", conditions]']



u"""
The ``configure`` subcommand generates a stub configuration for
the specified class (or for the top level configuration if no class specified).

.. code-block:: bash

    $ allennlp configure --help
    usage: allennlp configure [-h] [class]

    Generate a configuration stub for a specific class (or for config as a whole if [class] is omitted).

    positional arguments:
    class

    optional arguments:
    -h, --help            show this help message and exit
"""


from __future__ import absolute_import
from __future__ import print_function
import argparse

from allennlp.commands.subcommand import Subcommand
from allennlp.common.configuration import configure, Config, render_config

class Configure(Subcommand):
    def add_subparser(self, name     , parser                            )                           :
        # pylint: disable=protected-access
        description = u'''Generate a configuration stub for a specific class (or for config as a whole)'''
        subparser = parser.add_parser(
                name, description=description, help=u'Generate configuration stubs.')

        subparser.add_argument(u'cla55', nargs=u'?', default=u'', metavar=u'class')
        subparser.set_defaults(func=_configure)

        return subparser

def _configure(args                    )        :
    cla55 = args.cla55
    parts = cla55.split(u".")
    module = u".".join(parts[:-1])
    class_name = parts[-1]

    print()

    try:
        config = configure(cla55)
        if isinstance(config, Config):
            if cla55:
                print("configuration stub for {cla55}:\n")
            else:
                print("configuration stub for AllenNLP:\n")
            print(render_config(config))
        else:
            print("{class_name} is an abstract base class, choose one of the following subclasses:\n")
            for subclass in config:
                print(u"\t", subclass)
    except ModuleNotFoundError:
        print("unable to load module {module}")
    except AttributeError:
        print("class {class_name} does not exist in module {module}")

    print()



# pylint: disable=no-self-use,invalid-name


from __future__ import division
from __future__ import absolute_import
from allennlp.common.testing import AllenNlpTestCase

from allennlp.models.archival import load_archive
from allennlp.predictors import Predictor


class TestCorefPredictor(AllenNlpTestCase):
    def test_uses_named_inputs(self):
        inputs = {u"document": u"This is a single string document about a test. Sometimes it "
                              u"contains coreferent parts."}
        archive = load_archive(self.FIXTURES_ROOT / u'coref' / u'serialization' / u'model.tar.gz')
        predictor = Predictor.from_archive(archive, u'coreference-resolution')

        result = predictor.predict_json(inputs)

        document = result[u"document"]
        assert document == [u'This', u'is', u'a', u'single', u'string',
                            u'document', u'about', u'a', u'test', u'.', u'Sometimes',
                            u'it', u'contains', u'coreferent', u'parts', u'.']

        clusters = result[u"clusters"]
        assert isinstance(clusters, list)
        for cluster in clusters:
            assert isinstance(cluster, list)
            for mention in cluster:
                # Spans should be integer indices.
                assert isinstance(mention[0], int)
                assert isinstance(mention[1], int)
                # Spans should be inside document.
                assert 0 < mention[0] <= len(document)
                assert 0 < mention[1] <= len(document)





from __future__ import division
from __future__ import absolute_import
#overrides

import torch

from allennlp.modules.seq2vec_encoders.seq2vec_encoder import Seq2VecEncoder
from allennlp.nn.util import get_lengths_from_binary_sequence_mask

class BagOfEmbeddingsEncoder(Seq2VecEncoder):
    u"""
    A ``BagOfEmbeddingsEncoder`` is a simple :class:`Seq2VecEncoder` which simply sums the embeddings
    of a sequence across the time dimension. The input to this module is of shape ``(batch_size, num_tokens,
    embedding_dim)``, and the output is of shape ``(batch_size, embedding_dim)``.

    Parameters
    ----------
    embedding_dim: ``int``
        This is the input dimension to the encoder.
    averaged: ``bool``, optional (default=``False``)
        If ``True``, this module will average the embeddings across time, rather than simply summing
        (ie. we will divide the summed embeddings by the length of the sentence).
    """
    def __init__(self,
                 embedding_dim     ,
                 averaged       = False)        :
        super(BagOfEmbeddingsEncoder, self).__init__()
        self._embedding_dim = embedding_dim
        self._averaged = averaged

    #overrides
    def get_input_dim(self)       :
        return self._embedding_dim

    #overrides
    def get_output_dim(self)       :
        return self._embedding_dim

    def forward(self, tokens              , mask               = None):  #pylint: disable=arguments-differ
        if mask is not None:
            tokens = tokens * mask.unsqueeze(-1).float()

        # Our input has shape `(batch_size, num_tokens, embedding_dim)`, so we sum out the `num_tokens`
        # dimension.
        summed = tokens.sum(1)

        if self._averaged:
            if mask is not None:
                lengths = get_lengths_from_binary_sequence_mask(mask)
                length_mask = (lengths > 0)

                # Set any length 0 to 1, to avoid dividing by zero.
                lengths = torch.max(lengths, lengths.new_ones(1))
            else:
                lengths = tokens.new_full((1,), fill_value=tokens.size(1))
                length_mask = None

            summed = summed / lengths.unsqueeze(-1).float()

            if length_mask is not None:
                summed = summed * (length_mask > 0).float().unsqueeze(-1)

        return summed

BagOfEmbeddingsEncoder = Seq2VecEncoder.register(u"boe")(BagOfEmbeddingsEncoder)



# pylint: disable=no-self-use,invalid-name,protected-access


from __future__ import division
from __future__ import absolute_import
import os
import pytest

from allennlp.common.testing import AllenNlpTestCase
from allennlp.training.metrics import WikiTablesAccuracy
from allennlp.training.metrics.wikitables_accuracy import SEMPRE_ABBREVIATIONS_PATH, SEMPRE_GRAMMAR_PATH

class WikiTablesAccuracyTest(AllenNlpTestCase):
    def setUp(self):
        super(WikiTablesAccuracyTest, self).setUp()
        self.should_remove_sempre_abbreviations = not os.path.exists(SEMPRE_ABBREVIATIONS_PATH)
        self.should_remove_sempre_grammar = not os.path.exists(SEMPRE_GRAMMAR_PATH)

    def tearDown(self):
        super(WikiTablesAccuracyTest, self).tearDown()
        if self.should_remove_sempre_abbreviations and os.path.exists(SEMPRE_ABBREVIATIONS_PATH):
            os.remove(SEMPRE_ABBREVIATIONS_PATH)
        if self.should_remove_sempre_grammar and os.path.exists(SEMPRE_GRAMMAR_PATH):
            os.remove(SEMPRE_GRAMMAR_PATH)

    def test_accuracy_is_scored_correctly(self):
        # This is the first example in our test fixture.
        example_string = (u'(example (id nt-0) (utterance "what was the last year where this team '
                          u'was a part of the usl a-league?") (context (graph '
                          u'tables.TableKnowledgeGraph tables/590.csv)) '
                          u'(targetValue (list (description "2004"))))')

        # This logical form should produce the correct denotation (the "targetValue" above) given
        # the table.
        logical_form = (u'((reverse fb:row.row.year) (fb:row.row.index (max '
                        u'((reverse fb:row.row.index) (fb:row.row.league fb:cell.usl_a_league)))))')
        wikitables_accuracy = WikiTablesAccuracy(table_directory=unicode(self.FIXTURES_ROOT / u'data' / u'wikitables/'))
        wikitables_accuracy(logical_form, example_string)
        assert wikitables_accuracy._count == 1
        assert wikitables_accuracy._correct == 1

        # Testing that we handle bad logical forms correctly.
        wikitables_accuracy(None, example_string)
        assert wikitables_accuracy._count == 2
        assert wikitables_accuracy._correct == 1

        wikitables_accuracy(u'Error producing logical form', example_string)
        assert wikitables_accuracy._count == 3
        assert wikitables_accuracy._correct == 1

        # And an incorrect logical form.
        wikitables_accuracy(u'(fb:row.row.league fb:cell.3rd_usl_3rd)', example_string)
        assert wikitables_accuracy._count == 4
        assert wikitables_accuracy._correct == 1

WikiTablesAccuracyTest = pytest.mark.java(WikiTablesAccuracyTest)




from __future__ import absolute_import
#typing
#overrides

import torch
from torch.nn.modules.linear import Linear
from nltk import Tree

from allennlp.common.checks import check_dimensions_match
from allennlp.data import Vocabulary
from allennlp.modules import Seq2SeqEncoder, TimeDistributed, TextFieldEmbedder, FeedForward
from allennlp.modules.token_embedders import Embedding
from allennlp.modules.span_extractors.span_extractor import SpanExtractor
from allennlp.models.model import Model
from allennlp.nn import InitializerApplicator, RegularizerApplicator
from allennlp.nn.util import get_text_field_mask, sequence_cross_entropy_with_logits
from allennlp.nn.util import last_dim_softmax, get_lengths_from_binary_sequence_mask
from allennlp.training.metrics import CategoricalAccuracy
from allennlp.training.metrics import EvalbBracketingScorer, DEFAULT_EVALB_DIR
from allennlp.common.checks import ConfigurationError
try:
    from itertools import izip
except:
    izip = zip

class SpanInformation(NamedTuple):
    u"""
    A helper namedtuple for handling decoding information.

    Parameters
    ----------
    start : ``int``
        The start index of the span.
    end : ``int``
        The exclusive end index of the span.
    no_label_prob : ``float``
        The probability of this span being assigned the ``NO-LABEL`` label.
    label_prob : ``float``
        The probability of the most likely label.
    """
    #tart: int
    #nd: int
    #abel_prob: float
    #o_label_prob: float
    #abel_index: int


class SpanConstituencyParser(Model):
    u"""
    This ``SpanConstituencyParser`` simply encodes a sequence of text
    with a stacked ``Seq2SeqEncoder``, extracts span representations using a
    ``SpanExtractor``, and then predicts a label for each span in the sequence.
    These labels are non-terminal nodes in a constituency parse tree, which we then
    greedily reconstruct.

    Parameters
    ----------
    vocab : ``Vocabulary``, required
        A Vocabulary, required in order to compute sizes for input/output projections.
    text_field_embedder : ``TextFieldEmbedder``, required
        Used to embed the ``tokens`` ``TextField`` we get as input to the model.
    span_extractor : ``SpanExtractor``, required.
        The method used to extract the spans from the encoded sequence.
    encoder : ``Seq2SeqEncoder``, required.
        The encoder that we will use in between embedding tokens and
        generating span representations.
    feedforward : ``FeedForward``, required.
        The FeedForward layer that we will use in between the encoder and the linear
        projection to a distribution over span labels.
    pos_tag_embedding : ``Embedding``, optional.
        Used to embed the ``pos_tags`` ``SequenceLabelField`` we get as input to the model.
    initializer : ``InitializerApplicator``, optional (default=``InitializerApplicator()``)
        Used to initialize the model parameters.
    regularizer : ``RegularizerApplicator``, optional (default=``None``)
        If provided, will be used to calculate the regularization penalty during training.
    evalb_directory_path : ``str``, optional (default=``DEFAULT_EVALB_DIR``)
        The path to the directory containing the EVALB executable used to score
        bracketed parses. By default, will use the EVALB included with allennlp,
        which is located at allennlp/tools/EVALB . If ``None``, EVALB scoring
        is not used.
    """
    def __init__(self,
                 vocab            ,
                 text_field_embedder                   ,
                 span_extractor               ,
                 encoder                ,
                 feedforward              = None,
                 pos_tag_embedding            = None,
                 initializer                        = InitializerApplicator(),
                 regularizer                                  = None,
                 evalb_directory_path      = DEFAULT_EVALB_DIR)        :
        super(SpanConstituencyParser, self).__init__(vocab, regularizer)

        self.text_field_embedder = text_field_embedder
        self.span_extractor = span_extractor
        self.num_classes = self.vocab.get_vocab_size(u"labels")
        self.encoder = encoder
        self.feedforward_layer = TimeDistributed(feedforward) if feedforward else None
        self.pos_tag_embedding = pos_tag_embedding or None
        if feedforward is not None:
            output_dim = feedforward.get_output_dim()
        else:
            output_dim = span_extractor.get_output_dim()

        self.tag_projection_layer = TimeDistributed(Linear(output_dim, self.num_classes))

        representation_dim = text_field_embedder.get_output_dim()
        if pos_tag_embedding is not None:
            representation_dim += pos_tag_embedding.get_output_dim()
        check_dimensions_match(representation_dim,
                               encoder.get_input_dim(),
                               u"representation dim (tokens + optional POS tags)",
                               u"encoder input dim")
        check_dimensions_match(encoder.get_output_dim(),
                               span_extractor.get_input_dim(),
                               u"encoder input dim",
                               u"span extractor input dim")
        if feedforward is not None:
            check_dimensions_match(span_extractor.get_output_dim(),
                                   feedforward.get_input_dim(),
                                   u"span extractor output dim",
                                   u"feedforward input dim")

        self.tag_accuracy = CategoricalAccuracy()

        if evalb_directory_path is not None:
            self._evalb_score = EvalbBracketingScorer(evalb_directory_path)
        else:
            self._evalb_score = None
        initializer(self)

    #overrides
    def forward(self,  # type: ignore
                tokens                             ,
                spans                  ,
                metadata                      ,
                pos_tags                              = None,
                span_labels                   = None)                           :
        # pylint: disable=arguments-differ
        u"""
        Parameters
        ----------
        tokens : Dict[str, torch.LongTensor], required
            The output of ``TextField.as_array()``, which should typically be passed directly to a
            ``TextFieldEmbedder``. This output is a dictionary mapping keys to ``TokenIndexer``
            tensors.  At its most basic, using a ``SingleIdTokenIndexer`` this is: ``{"tokens":
            Tensor(batch_size, num_tokens)}``. This dictionary will have the same keys as were used
            for the ``TokenIndexers`` when you created the ``TextField`` representing your
            sequence.  The dictionary is designed to be passed directly to a ``TextFieldEmbedder``,
            which knows how to combine different word representations into a single vector per
            token in your input.
        spans : ``torch.LongTensor``, required.
            A tensor of shape ``(batch_size, num_spans, 2)`` representing the
            inclusive start and end indices of all possible spans in the sentence.
        metadata : List[Dict[str, Any]], required.
            A dictionary of metadata for each batch element which has keys:
                tokens : ``List[str]``, required.
                    The original string tokens in the sentence.
                gold_tree : ``nltk.Tree``, optional (default = None)
                    Gold NLTK trees for use in evaluation.
                pos_tags : ``List[str]``, optional.
                    The POS tags for the sentence. These can be used in the
                    model as embedded features, but they are passed here
                    in addition for use in constructing the tree.
        pos_tags : ``torch.LongTensor``, optional (default = None)
            The output of a ``SequenceLabelField`` containing POS tags.
        span_labels : ``torch.LongTensor``, optional (default = None)
            A torch tensor representing the integer gold class labels for all possible
            spans, of shape ``(batch_size, num_spans)``.

        Returns
        -------
        An output dictionary consisting of:
        class_probabilities : ``torch.FloatTensor``
            A tensor of shape ``(batch_size, num_spans, span_label_vocab_size)``
            representing a distribution over the label classes per span.
        spans : ``torch.LongTensor``
            The original spans tensor.
        tokens : ``List[List[str]]``, required.
            A list of tokens in the sentence for each element in the batch.
        pos_tags : ``List[List[str]]``, required.
            A list of POS tags in the sentence for each element in the batch.
        num_spans : ``torch.LongTensor``, required.
            A tensor of shape (batch_size), representing the lengths of non-padded spans
            in ``enumerated_spans``.
        loss : ``torch.FloatTensor``, optional
            A scalar loss to be optimised.
        """
        embedded_text_input = self.text_field_embedder(tokens)
        if pos_tags is not None and self.pos_tag_embedding is not None:
            embedded_pos_tags = self.pos_tag_embedding(pos_tags)
            embedded_text_input = torch.cat([embedded_text_input, embedded_pos_tags], -1)
        elif self.pos_tag_embedding is not None:
            raise ConfigurationError(u"Model uses a POS embedding, but no POS tags were passed.")

        mask = get_text_field_mask(tokens)
        # Looking at the span start index is enough to know if
        # this is padding or not. Shape: (batch_size, num_spans)
        span_mask = (spans[:, :, 0] >= 0).squeeze(-1).long()
        if span_mask.dim() == 1:
            # This happens if you use batch_size 1 and encounter
            # a length 1 sentence in PTB, which do exist. -.-
            span_mask = span_mask.unsqueeze(-1)
        if span_labels is not None and span_labels.dim() == 1:
            span_labels = span_labels.unsqueeze(-1)

        num_spans = get_lengths_from_binary_sequence_mask(span_mask)

        encoded_text = self.encoder(embedded_text_input, mask)

        span_representations = self.span_extractor(encoded_text, spans, mask, span_mask)

        if self.feedforward_layer is not None:
            span_representations = self.feedforward_layer(span_representations)

        logits = self.tag_projection_layer(span_representations)
        class_probabilities = last_dim_softmax(logits, span_mask.unsqueeze(-1))

        output_dict = {
                u"class_probabilities": class_probabilities,
                u"spans": spans,
                u"tokens": [meta[u"tokens"] for meta in metadata],
                u"pos_tags": [meta.get(u"pos_tags") for meta in metadata],
                u"num_spans": num_spans
        }
        if span_labels is not None:
            loss = sequence_cross_entropy_with_logits(logits, span_labels, span_mask)
            self.tag_accuracy(class_probabilities, span_labels, span_mask)
            output_dict[u"loss"] = loss

        # The evalb score is expensive to compute, so we only compute
        # it for the validation and test sets.
        batch_gold_trees = [meta.get(u"gold_tree") for meta in metadata]
        if all(batch_gold_trees) and self._evalb_score is not None and not self.training:
            gold_pos_tags                  = [list(izip(*tree.pos()))[1]
                                              for tree in batch_gold_trees]
            predicted_trees = self.construct_trees(class_probabilities.cpu().data,
                                                   spans.cpu().data,
                                                   num_spans.data,
                                                   output_dict[u"tokens"],
                                                   gold_pos_tags)
            self._evalb_score(predicted_trees, batch_gold_trees)

        return output_dict

    #overrides
    def decode(self, output_dict                         )                           :
        u"""
        Constructs an NLTK ``Tree`` given the scored spans. We also switch to exclusive
        span ends when constructing the tree representation, because it makes indexing
        into lists cleaner for ranges of text, rather than individual indices.

        Finally, for batch prediction, we will have padded spans and class probabilities.
        In order to make this less confusing, we remove all the padded spans and
        distributions from ``spans`` and ``class_probabilities`` respectively.
        """
        all_predictions = output_dict[u'class_probabilities'].cpu().data
        all_spans = output_dict[u"spans"].cpu().data

        all_sentences = output_dict[u"tokens"]
        all_pos_tags = output_dict[u"pos_tags"] if all(output_dict[u"pos_tags"]) else None
        num_spans = output_dict[u"num_spans"].data
        trees = self.construct_trees(all_predictions, all_spans, num_spans, all_sentences, all_pos_tags)

        batch_size = all_predictions.size(0)
        output_dict[u"spans"] = [all_spans[i, :num_spans[i]] for i in range(batch_size)]
        output_dict[u"class_probabilities"] = [all_predictions[i, :num_spans[i], :] for i in range(batch_size)]

        output_dict[u"trees"] = trees
        return output_dict

    def construct_trees(self,
                        predictions                   ,
                        all_spans                  ,
                        num_spans                  ,
                        sentences                 ,
                        pos_tags                  = None)              :
        u"""
        Construct ``nltk.Tree``'s for each batch element by greedily nesting spans.
        The trees use exclusive end indices, which contrasts with how spans are
        represented in the rest of the model.

        Parameters
        ----------
        predictions : ``torch.FloatTensor``, required.
            A tensor of shape ``(batch_size, num_spans, span_label_vocab_size)``
            representing a distribution over the label classes per span.
        all_spans : ``torch.LongTensor``, required.
            A tensor of shape (batch_size, num_spans, 2), representing the span
            indices we scored.
        num_spans : ``torch.LongTensor``, required.
            A tensor of shape (batch_size), representing the lengths of non-padded spans
            in ``enumerated_spans``.
        sentences : ``List[List[str]]``, required.
            A list of tokens in the sentence for each element in the batch.
        pos_tags : ``List[List[str]]``, optional (default = None).
            A list of POS tags for each word in the sentence for each element
            in the batch.

        Returns
        -------
        A ``List[Tree]`` containing the decoded trees for each element in the batch.
        """
        # Switch to using exclusive end spans.
        exclusive_end_spans = all_spans.clone()
        exclusive_end_spans[:, :, -1] += 1
        no_label_id = self.vocab.get_token_index(u"NO-LABEL", u"labels")

        trees             = []
        for batch_index, (scored_spans, spans, sentence) in enumerate(izip(predictions,
                                                                          exclusive_end_spans,
                                                                          sentences)):
            selected_spans = []
            for prediction, span in izip(scored_spans[:num_spans[batch_index]],
                                        spans[:num_spans[batch_index]]):
                start, end = span
                no_label_prob = prediction[no_label_id]
                label_prob, label_index = torch.max(prediction, -1)

                # Does the span have a label != NO-LABEL or is it the root node?
                # If so, include it in the spans that we consider.
                if int(label_index) != no_label_id or (start == 0 and end == len(sentence)):
                    # TODO(Mark): Remove this once pylint sorts out named tuples.
                    # https://github.com/PyCQA/pylint/issues/1418
                    selected_spans.append(SpanInformation(start=int(start), # pylint: disable=no-value-for-parameter
                                                          end=int(end),
                                                          label_prob=float(label_prob),
                                                          no_label_prob=float(no_label_prob),
                                                          label_index=int(label_index)))

            # The spans we've selected might overlap, which causes problems when we try
            # to construct the tree as they won't nest properly.
            consistent_spans = self.resolve_overlap_conflicts_greedily(selected_spans)

            spans_to_labels = dict(((span.start, span.end), self.vocab.get_token_from_index(span.label_index, u"labels"))
                               for span in consistent_spans)
            sentence_pos = pos_tags[batch_index] if pos_tags is not None else None
            trees.append(self.construct_tree_from_spans(spans_to_labels, sentence, sentence_pos))

        return trees

    @staticmethod
    def resolve_overlap_conflicts_greedily(spans                       )                         :
        u"""
        Given a set of spans, removes spans which overlap by evaluating the difference
        in probability between one being labeled and the other explicitly having no label
        and vice-versa. The worst case time complexity of this method is ``O(k * n^4)`` where ``n``
        is the length of the sentence that the spans were enumerated from (and therefore
        ``k * m^2`` complexity with respect to the number of spans ``m``) and ``k`` is the
        number of conflicts. However, in practice, there are very few conflicts. Hopefully.

        This function modifies ``spans`` to remove overlapping spans.

        Parameters
        ----------
        spans: ``List[SpanInformation]``, required.
            A list of spans, where each span is a ``namedtuple`` containing the
            following attributes:

        start : ``int``
            The start index of the span.
        end : ``int``
            The exclusive end index of the span.
        no_label_prob : ``float``
            The probability of this span being assigned the ``NO-LABEL`` label.
        label_prob : ``float``
            The probability of the most likely label.

        Returns
        -------
        A modified list of ``spans``, with the conflicts resolved by considering local
        differences between pairs of spans and removing one of the two spans.
        """
        conflicts_exist = True
        while conflicts_exist:
            conflicts_exist = False
            for span1_index, span1 in enumerate(spans):
                for span2_index, span2 in list(enumerate(spans))[span1_index + 1:]:
                    if (span1.start < span2.start < span1.end < span2.end or
                                span2.start < span1.start < span2.end < span1.end):
                        # The spans overlap.
                        conflicts_exist = True
                        # What's the more likely situation: that span2 was labeled
                        # and span1 was unlabled, or that span1 was labeled and span2
                        # was unlabled? In the first case, we delete span2 from the
                        # set of spans to form the tree - in the second case, we delete
                        # span1.
                        if (span1.no_label_prob + span2.label_prob <
                                    span2.no_label_prob + span1.label_prob):
                            spans.pop(span2_index)
                        else:
                            spans.pop(span1_index)
                        break
        return spans

    @staticmethod
    def construct_tree_from_spans(spans_to_labels                            ,
                                  sentence           ,
                                  pos_tags            = None)        :
        u"""
        Parameters
        ----------
        spans_to_labels : ``Dict[Tuple[int, int], str]``, required.
            A mapping from spans to constituency labels.
        sentence : ``List[str]``, required.
            A list of tokens forming the sentence to be parsed.
        pos_tags : ``List[str]``, optional (default = None)
            A list of the pos tags for the words in the sentence, if they
            were either predicted or taken as input to the model.

        Returns
        -------
        An ``nltk.Tree`` constructed from the labelled spans.
        """
        def assemble_subtree(start     , end     ):
            if (start, end) in spans_to_labels:
                # Some labels contain nested spans, e.g S-VP.
                # We actually want to create (S (VP ...)) nodes
                # for these labels, so we split them up here.
                labels            = spans_to_labels[(start, end)].split(u"-")
            else:
                labels = None

            # This node is a leaf.
            if end - start == 1:
                word = sentence[start]
                pos_tag = pos_tags[start] if pos_tags is not None else u"XX"
                tree = Tree(pos_tag, [word])
                if labels is not None and pos_tags is not None:
                    # If POS tags were passed explicitly,
                    # they are added as pre-terminal nodes.
                    while labels:
                        tree = Tree(labels.pop(), [tree])
                elif labels is not None:
                    # Otherwise, we didn't want POS tags
                    # at all.
                    tree = Tree(labels.pop(), [word])
                    while labels:
                        tree = Tree(labels.pop(), [tree])
                return [tree]

            argmax_split = start + 1
            # Find the next largest subspan such that
            # the left hand side is a constituent.
            for split in range(end - 1, start, -1):
                if (start, split) in spans_to_labels:
                    argmax_split = split
                    break

            left_trees = assemble_subtree(start, argmax_split)
            right_trees = assemble_subtree(argmax_split, end)
            children = left_trees + right_trees
            if labels is not None:
                while labels:
                    children = [Tree(labels.pop(), children)]
            return children

        tree = assemble_subtree(0, len(sentence))
        return tree[0]

    #overrides
    def get_metrics(self, reset       = False)                    :
        all_metrics = {}
        all_metrics[u"tag_accuracy"] = self.tag_accuracy.get_metric(reset=reset)
        if self._evalb_score is not None:
            evalb_metrics = self._evalb_score.get_metric(reset=reset)
            all_metrics.update(evalb_metrics)
        return all_metrics

SpanConstituencyParser = Model.register(u"constituency_parser")(SpanConstituencyParser)





from __future__ import absolute_import
import torch

#overrides
from allennlp.modules.attention.attention import Attention
from allennlp.modules.similarity_functions import DotProductSimilarity, SimilarityFunction


class LegacyAttention(Attention):
    u"""
    Computes attention between a vector and a matrix using a similarity function.
    This should be considered deprecated, as it consumes more memory than the specialized attention modules.
    """

    def __init__(self,
                 similarity_function                     = None,
                 normalize       = True)        :
        super(LegacyAttention, self).__init__(normalize)
        self._similarity_function = similarity_function or DotProductSimilarity()

    #overrides
    def _forward_internal(self, vector              , matrix              )                :
        tiled_vector = vector.unsqueeze(1).expand(vector.size()[0],
                                                  matrix.size()[1],
                                                  vector.size()[1])
        return self._similarity_function(tiled_vector, matrix)

LegacyAttention = Attention.register(u"legacy")(LegacyAttention)



u"""
The various :class:`~allennlp.data.iterators.data_iterator.DataIterator` subclasses
can be used to iterate over datasets with different batching and padding schemes.
"""


from __future__ import absolute_import
from allennlp.data.iterators.data_iterator import DataIterator
from allennlp.data.iterators.basic_iterator import BasicIterator
from allennlp.data.iterators.bucket_iterator import BucketIterator
from allennlp.data.iterators.epoch_tracking_bucket_iterator import EpochTrackingBucketIterator




from __future__ import absolute_import
#typing

import torch

from allennlp.data.vocabulary import Vocabulary



class Field:
    u"""
    A ``Field`` is some piece of a data instance that ends up as an tensor in a model (either as an
    input or an output).  Data instances are just collections of fields.

    Fields go through up to two steps of processing: (1) tokenized fields are converted into token
    ids, (2) fields containing token ids (or any other numeric data) are padded (if necessary) and
    converted into tensors.  The ``Field`` API has methods around both of these steps, though they
    may not be needed for some concrete ``Field`` classes - if your field doesn't have any strings
    that need indexing, you don't need to implement ``count_vocab_items`` or ``index``.  These
    methods ``pass`` by default.

    Once a vocabulary is computed and all fields are indexed, we will determine padding lengths,
    then intelligently batch together instances and pad them into actual tensors.
    """
    def count_vocab_items(self, counter                           ):
        u"""
        If there are strings in this field that need to be converted into integers through a
        :class:`Vocabulary`, here is where we count them, to determine which tokens are in or out
        of the vocabulary.

        If your ``Field`` does not have any strings that need to be converted into indices, you do
        not need to implement this method.

        A note on this ``counter``: because ``Fields`` can represent conceptually different things,
        we separate the vocabulary items by `namespaces`.  This way, we can use a single shared
        mechanism to handle all mappings from strings to integers in all fields, while keeping
        words in a ``TextField`` from sharing the same ids with labels in a ``LabelField`` (e.g.,
        "entailment" or "contradiction" are labels in an entailment task)

        Additionally, a single ``Field`` might want to use multiple namespaces - ``TextFields`` can
        be represented as a combination of word ids and character ids, and you don't want words and
        characters to share the same vocabulary - "a" as a word should get a different id from "a"
        as a character, and the vocabulary sizes of words and characters are very different.

        Because of this, the first key in the ``counter`` object is a `namespace`, like "tokens",
        "token_characters", "tags", or "labels", and the second key is the actual vocabulary item.
        """
        pass

    def index(self, vocab            ):
        u"""
        Given a :class:`Vocabulary`, converts all strings in this field into (typically) integers.
        This `modifies` the ``Field`` object, it does not return anything.

        If your ``Field`` does not have any strings that need to be converted into indices, you do
        not need to implement this method.
        """
        pass

    def get_padding_lengths(self)                  :
        u"""
        If there are things in this field that need padding, note them here.  In order to pad a
        batch of instance, we get all of the lengths from the batch, take the max, and pad
        everything to that length (or use a pre-specified maximum length).  The return value is a
        dictionary mapping keys to lengths, like {'num_tokens': 13}.

        This is always called after :func:`index`.
        """
        raise NotImplementedError

    def as_tensor(self,
                  padding_lengths                ,
                  cuda_device      = -1)             :
        u"""
        Given a set of specified padding lengths, actually pad the data in this field and return a
        torch Tensor (or a more complex data structure) of the correct shape.  We also take a
        couple of parameters that are important when constructing torch Tensors.

        Parameters
        ----------
        padding_lengths : ``Dict[str, int]``
            This dictionary will have the same keys that were produced in
            :func:`get_padding_lengths`.  The values specify the lengths to use when padding each
            relevant dimension, aggregated across all instances in a batch.
        cuda_device : ``int``
            If cuda_device >= 0, GPUs are available and Pytorch was compiled with CUDA support, we
            will allocate tensors on this device instead of on the CPU.
        """
        raise NotImplementedError

    def empty_field(self)           :
        u"""
        So that ``ListField`` can pad the number of fields in a list (e.g., the number of answer
        option ``TextFields``), we need a representation of an empty field of each type.  This
        returns that.  This will only ever be called when we're to the point of calling
        :func:`as_tensor`, so you don't need to worry about ``get_padding_lengths``,
        ``count_vocab_items``, etc., being called on this empty field.

        We make this an instance method instead of a static method so that if there is any state
        in the Field, we can copy it over (e.g., the token indexers in ``TextField``).
        """
        raise NotImplementedError

    def batch_tensors(self, tensor_list                 )             :  # type: ignore
        u"""
        Takes the output of ``Field.as_tensor()`` from a list of ``Instances`` and merges it into
        one batched tensor for this ``Field``.  The default implementation here in the base class
        handles cases where ``as_tensor`` returns a single torch tensor per instance.  If your
        subclass returns something other than this, you need to override this method.

        This operation does not modify ``self``, but in some cases we need the information
        contained in ``self`` in order to perform the batching, so this is an instance method, not
        a class method.
        """
        # pylint: disable=no-self-use
        return torch.stack(tensor_list)



# pylint: disable=no-self-use,invalid-name

from __future__ import absolute_import
from allennlp.common.testing import AllenNlpTestCase
from allennlp.semparse import util


class TestSemparseUtil(AllenNlpTestCase):
    def test_lisp_to_nested_expression(self):
        logical_form = u"((reverse fb:row.row.year) (fb:row.row.league fb:cell.usl_a_league))"
        expression = util.lisp_to_nested_expression(logical_form)
        assert expression == [[[u'reverse', u'fb:row.row.year'], [u'fb:row.row.league', u'fb:cell.usl_a_league']]]
        logical_form = u"(count (and (division 1) (tier (!= null))))"
        expression = util.lisp_to_nested_expression(logical_form)
        assert expression == [[u'count', [u'and', [u'division', u'1'], [u'tier', [u'!=', u'null']]]]]




from __future__ import absolute_import
import logging
import collections
#typing

#overrides

from allennlp.common.file_utils import cached_path
from allennlp.data.dataset_readers.dataset_reader import DatasetReader
from allennlp.data.fields import Field, ListField, TextField, SpanField, MetadataField, SequenceLabelField
from allennlp.data.instance import Instance
from allennlp.data.tokenizers import Token
from allennlp.data.token_indexers import SingleIdTokenIndexer, TokenIndexer
from allennlp.data.dataset_readers.dataset_utils import Ontonotes, enumerate_spans

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def canonicalize_clusters(clusters                                         )                               :
    u"""
    The CONLL 2012 data includes 2 annotatated spans which are identical,
    but have different ids. This checks all clusters for spans which are
    identical, and if it finds any, merges the clusters containing the
    identical spans.
    """
    merged_clusters                             = []
    for cluster in list(clusters.values()):
        cluster_with_overlapping_mention = None
        for mention in cluster:
            # Look at clusters we have already processed to
            # see if they contain a mention in the current
            # cluster for comparison.
            for cluster2 in merged_clusters:
                if mention in cluster2:
                    # first cluster in merged clusters
                    # which contains this mention.
                    cluster_with_overlapping_mention = cluster2
                    break
            # Already encountered overlap - no need to keep looking.
            if cluster_with_overlapping_mention is not None:
                break
        if cluster_with_overlapping_mention is not None:
            # Merge cluster we are currently processing into
            # the cluster in the processed list.
            cluster_with_overlapping_mention.update(cluster)
        else:
            merged_clusters.append(set(cluster))
    return [list(c) for c in merged_clusters]


class ConllCorefReader(DatasetReader):
    u"""
    Reads a single CoNLL-formatted file. This is the same file format as used in the
    :class:`~allennlp.data.dataset_readers.semantic_role_labelling.SrlReader`, but is preprocessed
    to dump all documents into a single file per train, dev and test split. See
    scripts/compile_coref_data.sh for more details of how to pre-process the Ontonotes 5.0 data
    into the correct format.

    Returns a ``Dataset`` where the ``Instances`` have four fields: ``text``, a ``TextField``
    containing the full document text, ``spans``, a ``ListField[SpanField]`` of inclusive start and
    end indices for span candidates, and ``metadata``, a ``MetadataField`` that stores the instance's
    original text. For data with gold cluster labels, we also include the original ``clusters``
    (a list of list of index pairs) and a ``SequenceLabelField`` of cluster ids for every span
    candidate.

    Parameters
    ----------
    max_span_width: ``int``, required.
        The maximum width of candidate spans to consider.
    token_indexers : ``Dict[str, TokenIndexer]``, optional
        This is used to index the words in the document.  See :class:`TokenIndexer`.
        Default is ``{"tokens": SingleIdTokenIndexer()}``.
    """
    def __init__(self,
                 max_span_width     ,
                 token_indexers                          = None,
                 lazy       = False)        :
        super(ConllCorefReader, self).__init__(lazy)
        self._max_span_width = max_span_width
        self._token_indexers = token_indexers or {u"tokens": SingleIdTokenIndexer()}

    #overrides
    def _read(self, file_path     ):
        # if `file_path` is a URL, redirect to the cache
        file_path = cached_path(file_path)

        ontonotes_reader = Ontonotes()
        for sentences in ontonotes_reader.dataset_document_iterator(file_path):
            clusters                                          = collections.defaultdict(list)

            total_tokens = 0
            for sentence in sentences:
                for typed_span in sentence.coref_spans:
                    # Coref annotations are on a _per sentence_
                    # basis, so we need to adjust them to be relative
                    # to the length of the document.
                    span_id, (start, end) = typed_span
                    clusters[span_id].append((start + total_tokens,
                                              end + total_tokens))
                total_tokens += len(sentence.words)

            canonical_clusters = canonicalize_clusters(clusters)
            yield self.text_to_instance([s.words for s in sentences], canonical_clusters)

    #overrides
    def text_to_instance(self,  # type: ignore
                         sentences                 ,
                         gold_clusters                                        = None)            :
        # pylint: disable=arguments-differ
        u"""
        Parameters
        ----------
        sentences : ``List[List[str]]``, required.
            A list of lists representing the tokenised words and sentences in the document.
        gold_clusters : ``Optional[List[List[Tuple[int, int]]]]``, optional (default = None)
            A list of all clusters in the document, represented as word spans. Each cluster
            contains some number of spans, which can be nested and overlap, but will never
            exactly match between clusters.

        Returns
        -------
        An ``Instance`` containing the following ``Fields``:
            text : ``TextField``
                The text of the full document.
            spans : ``ListField[SpanField]``
                A ListField containing the spans represented as ``SpanFields``
                with respect to the document text.
            span_labels : ``SequenceLabelField``, optional
                The id of the cluster which each possible span belongs to, or -1 if it does
                 not belong to a cluster. As these labels have variable length (it depends on
                 how many spans we are considering), we represent this a as a ``SequenceLabelField``
                 with respect to the ``spans ``ListField``.
        """
        flattened_sentences = [self._normalize_word(word)
                               for sentence in sentences
                               for word in sentence]

        metadata                 = {u"original_text": flattened_sentences}
        if gold_clusters is not None:
            metadata[u"clusters"] = gold_clusters

        text_field = TextField([Token(word) for word in flattened_sentences], self._token_indexers)

        cluster_dict = {}
        if gold_clusters is not None:
            for cluster_id, cluster in enumerate(gold_clusters):
                for mention in cluster:
                    cluster_dict[tuple(mention)] = cluster_id

        spans              = []
        span_labels                      = [] if gold_clusters is not None else None

        sentence_offset = 0
        for sentence in sentences:
            for start, end in enumerate_spans(sentence,
                                              offset=sentence_offset,
                                              max_span_width=self._max_span_width):
                if span_labels is not None:
                    if (start, end) in cluster_dict:
                        span_labels.append(cluster_dict[(start, end)])
                    else:
                        span_labels.append(-1)

                spans.append(SpanField(start, end, text_field))
            sentence_offset += len(sentence)

        span_field = ListField(spans)
        metadata_field = MetadataField(metadata)

        fields                   = {u"text": text_field,
                                    u"spans": span_field,
                                    u"metadata": metadata_field}
        if span_labels is not None:
            fields[u"span_labels"] = SequenceLabelField(span_labels, span_field)

        return Instance(fields)

    @staticmethod
    def _normalize_word(word):
        if word == u"/." or word == u"/?":
            return word[1:]
        else:
            return word

ConllCorefReader = DatasetReader.register(u"coref")(ConllCorefReader)
